#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import Tag, Text
from wdom.document import get_document


class App(Tag):
    tag = 'sci-app'


class H1(Tag):
    tag = 'h1'

class Div(Tag):
    tag = 'div'


def sample_page() -> Tag:
    app = App()
    text = H1(parent=app)
    text.textContent = 'Click!'
    div = Div(parent=app)
    count = 0

    def insert(data):
        nonlocal count
        count += 1
        child = Div()
        child.textContent = 'count {}'.format(str(count))
        if not div.hasChildNodes():
            div.appendChild(child)
        else:
            div.insertBefore(child, div.firstChild)

    text.addEventListener('click', insert)

    page = get_document(app=app)

    return page




#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'KATHAMO'
project_url = 'http://kathamo.github.io/'
project_repository = 'https://github.com/kathamo/Kathamo'
license = 'MIT License'
license_url = 'https://github.com/kathamo/Kathamo/blob/master/LICENSE'

css_files = [
    '_static/css/kathamo.min.css',
]

class Button(Button):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAttribute('data-role', 'button')


DefaultButton = NewTag('DefaultButton', 'button', DefaultButton)
PrimaryButton = NewTag('PrimaryButton', 'button', PrimaryButton)
SecondaryButton = NewTag('DefaultButton', 'button', SecondaryButton)
SuccessButton = NewTag('DefaultButton', 'button', SuccessButton)
InfoButton = NewTag('InfoButton', 'button', InfoButton)
WarningButton = NewTag('WarningButton', 'button', WarningButton)
DangerButton = NewTag('DangerButton', 'button', DangerButton)
ErrorButton = NewTag('ErrorButton', 'button', ErrorButton)
LinkButton = NewTag('LinkButton', 'button', LinkButton)

Row = NewTag('Row', 'div', Row, class_='container')
Col1 = NewTag('Col1', 'div', Col1, class_='col-sm-1 col-md-1 col-lg-1')
Col2 = NewTag('Col2', 'div', Col2, class_='col-sm-2 col-md-2 col-lg-2')
Col3 = NewTag('Col3', 'div', Col3, class_='col-sm-3 col-md-3 col-lg-3')
Col4 = NewTag('Col4', 'div', Col4, class_='col-sm-4 col-md-4 col-lg-4')
Col5 = NewTag('Col5', 'div', Col5, class_='col-sm-5 col-md-5 col-lg-5')
Col6 = NewTag('Col6', 'div', Col6, class_='col-sm-6 col-md-6 col-lg-6')
Col7 = NewTag('Col7', 'div', Col7, class_='col-sm-7 col-md-7 col-lg-7')
Col8 = NewTag('Col8', 'div', Col8, class_='col-sm-8 col-md-8 col-lg-8')
Col9 = NewTag('Col9', 'div', Col9, class_='col-sm-9 col-md-9 col-lg-9')
Col10 = NewTag('Col10', 'div', Col10, class_='col-sm-10 col-md-10 col-lg-10')
Col11 = NewTag('Col11', 'div', Col11, class_='col-sm-11 col-md-11 col-lg-11')

extended_classes = [
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    Row,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
from unittest.mock import MagicMock

from wdom import options
from wdom.interface import Event
from wdom.node import DocumentFragment, Comment, Text
from wdom.element import Attr, Element
from wdom.document import Document, get_document, get_new_document, set_document  # noqa
from wdom.document import getElementById, getElementByRimoId
from wdom.web_node import WebElement
from wdom.tag import Tag, HTMLElement, A
from wdom.testing import TestCase


class TestGetElement(TestCase):
    def setUp(self):
        super().setUp()
        Tag._elements_with_id.clear()
        Tag._elements_with_rimo_id.clear()
        self.doc = Document()

    def test_get_element_by_id(self):
        elm = Element(tag='a', id='a')
        self.assertIsNone(getElementById('a'))
        self.doc.appendChild(elm)
        self.assertIs(getElementById('a'), elm)

    def test_get_element_by_rimo_id(self):
        elm = WebElement(tag='a', id='a', rimo_id='b')
        self.assertIsNone(getElementById('a'))
        self.assertIsNone(getElementByRimoId('b'))
        self.doc.appendChild(elm)
        self.assertIs(getElementById('a'), elm)
        self.assertIsNone(getElementByRimoId('a'))
        self.assertIs(getElementByRimoId('b'), elm)


class TestMainDocument(TestCase):
    def setUp(self) -> None:
        super().setUp()
        Tag._elements_with_id.clear()
        Tag._elements_with_rimo_id.clear()
        self.doc = Document()
        self.doc.defaultView.customElements.reset()

    def tearDown(self):
        self.doc.defaultView.customElements.reset()
        super().tearDown()

    def test_blankpage(self) -> None:
        _re = re.compile(
            '\s*<!DOCTYPE html>'
            '\s*<html rimo_id="\d+">'
            '\s*<head rimo_id="\d+">'
            '\s*<meta( charset="utf-8"| rimo_id="\d+"){2}>'
            '\s*<title rimo_id="\d+">'
            '\s*W-DOM'
            '\s*</title>'
            '(\s*<script( type="text/javascript"| rimo_id="\d+"){2}>.*?</script>)?'
            '\s*</head>'
            '\s*<body rimo_id="\d+">'
            '\s*<script( type="text/javascript"| rimo_id="\d+"){2}>'
            '.*?</script>'
            '\s*</body>'
            '.*\s*</html>',
            re.S
        )
        html = self.doc.build()
        self.assertIsNotNone(_re.match(html))

    def test_get_element_by_id(self):
        self.assertIsNone(self.doc.getElementById('1'))
        elm = Element('a', parent=self.doc.body, id='a')
        self.assertIs(elm, self.doc.getElementById('a'))
        elm2 = Tag()
        self.assertIsNone(self.doc.getElementByRimoId(elm2.rimo_id))

    def test_get_element_by_remo_id(self):
        self.assertIsNone(self.doc.getElementByRimoId('1'))
        elm = Tag(parent=self.doc.body)
        self.assertIs(elm, self.doc.getElementByRimoId(elm.rimo_id))
        elm2 = Tag()
        self.assertIsNone(self.doc.getElementByRimoId(elm2.rimo_id))

    def test_add_jsfile(self) -> None:
        self.doc.add_jsfile('jsfile')
        _re = re.compile(
            '<body.*'
            '<script( src="jsfile"| type="text/javascript"| rimo_id="\d+"){3}'
            '>\s*</script>'
            '.*</body',
            re.S
        )
        self.assertIsNotNone(_re.search(self.doc.build()))

    def test_add_cssfile(self) -> None:
        self.doc.add_cssfile('cssfile')
        _re = re.compile(
            '<head rimo_id="\d+">.*'
            '<link( href="cssfile"| rel="stylesheet"| rimo_id="\d+"){3}>'
            '.*</head>'
            '', re.S
        )
        self.assertIsNotNone(_re.search(self.doc.build()))

    def test_add_header_link(self) -> None:
        self.doc.add_header('<link href="cssfile" rel="stylesheet">')
        self.assertIn(
            '<link href="cssfile" rel="stylesheet">',
            self.doc.build(),
        )

    def test_add_header_script(self) -> None:
        self.doc.add_header(
            '<script src="jsfile" type="text/javascript"></script>')
        self.assertIn(
            '<script src="jsfile" type="text/javascript"></script>',
            self.doc.build(),
        )

    def test_title(self) -> None:
        doc = Document(title='TEST')
        _re = re.compile(
            '<title rimo_id="\d+">\s*TEST\s*</title>',
            re.S
        )
        html = doc.build()
        self.assertIsNotNone(_re.search(html))
        self.assertNotIn('W-DOM', html)
        self.assertEqual('TEST', doc.title)

    def test_charset(self) -> None:
        doc = Document(charset='TEST')
        _re = re.compile(
            '<meta( charset="TEST"| rimo_id="\d+"){2}>',
            re.S
        )
        html = doc.build()
        self.assertIsNotNone(_re.search(html))
        self.assertNotIn('utf', html)
        self.assertEqual('TEST', doc.charset)

    def test_set_body(self) -> None:
        self.doc.body.prepend(Tag())
        html = self.doc.build()
        _re = re.compile(
            '<tag rimo_id="\d+">\s*</tag>',
            re.S
        )
        self.assertIsNotNone(_re.search(html))

    def test_set_body_string(self) -> None:
        string = 'testing'
        self.doc.body.prepend(string)
        html = self.doc.build()
        self.assertIn(string, html)

    def test_set_body_error(self) -> None:
        with self.assertRaises(TypeError):
            self.doc.body.prepend(1)

    def test_create_element(self):
        elm = self.doc.createElement('a')
        self.assertEqual(type(elm), A)
        self.assertRegex(elm.html, r'<a rimo_id="\d+"></a>')

    def test_create_element_unknown(self):
        elm = self.doc.createElement('aa')
        self.assertEqual(type(elm), HTMLElement)
        self.assertRegex(elm.html, r'<aa rimo_id="\d+"></aa>')

    def test_create_element_defclass(self):
        from wdom import element
        doc = Document(default_class=element.HTMLElement)
        elm = doc.createElement('a')
        self.assertEqual(type(elm), A)
        self.assertRegex(elm.html, '<a rimo_id="\d+"></a>')

    def test_create_element_defclass_unknown(self):
        from wdom import element
        doc = Document(default_class=element.HTMLElement)
        elm = doc.createElement('aa')
        self.assertEqual(type(elm), element.HTMLElement)
        self.assertRegex(elm.html, '<aa></aa>')

    def test_create_custom_element(self):
        class A(HTMLElement):
            pass
        self.doc.defaultView.customElements.define('a', A)
        elm = self.doc.createElement('a')
        self.assertEqual(type(elm), A)
        self.assertRegex(elm.html, '<a rimo_id="\d+"></a>')

    def test_create_custom_element_tag(self):
        class A(Tag):
            tag = 'a'
        self.doc.defaultView.customElements.define('a', A)
        elm = self.doc.createElement('a')
        self.assertEqual(type(elm), A)
        self.assertRegex(elm.html, '<a rimo_id="\d+"></a>')

    def test_create_documentfragment(self):
        df = self.doc.createDocumentFragment()
        self.assertEqual(type(df), DocumentFragment)
        self.assertFalse(df.hasChildNodes())

    def test_create_textnode(self):
        t = self.doc.createTextNode('text')
        self.assertEqual(type(t), Text)
        self.assertEqual(t.html, 'text')

    def test_create_comment(self):
        t = self.doc.createComment('text')
        self.assertEqual(type(t), Comment)
        self.assertEqual(t.html, '<!--text-->')

    def test_create_attr(self):
        a = self.doc.createAttribute('src')
        a.value = 'a'
        self.assertEqual(type(a), Attr)
        self.assertEqual(a.html, 'src="a"')
        tag = HTMLElement('tag')
        tag.setAttributeNode(a)
        self.assertRegex(tag.html, '<tag rimo_id="\d+" src="a"></tag>')

    def test_create_event(self):
        tag = HTMLElement('tag')
        mock = MagicMock(_is_coroutine=False)
        tag.addEventListener('a', mock)
        e = self.doc.createEvent('a')
        self.assertEqual(type(e), Event)
        tag.dispatchEvent(e)
        mock.assert_called_once_with(e)

    def test_custom_tag_theme_tag(self):
        from wdom import tag
        self.doc.register_theme(tag)
        elm = Tag(parent=self.doc.body)
        elm.innerHTML = '<div is="container"></div>'
        self.assertTrue(isinstance(elm.firstChild, tag.Container))

    def test_custom_tag_theme_default(self):
        from wdom.themes import default
        from wdom import tag
        self.doc.register_theme(default)
        elm = Tag(parent=self.doc.body)
        elm.innerHTML = '<div is="container"></div>'
        self.assertTrue(isinstance(elm.firstChild, default.Container))
        self.assertTrue(isinstance(elm.firstChild, tag.Container))

    def test_custom_tag_theme(self):
        from wdom.themes import bootstrap3
        from wdom import tag
        self.doc.register_theme(bootstrap3)
        elm = Tag(parent=self.doc.body)
        elm.innerHTML = '<div is="container"></div>'
        self.assertTrue(isinstance(elm.firstChild, bootstrap3.Container))
        self.assertTrue(isinstance(elm.firstChild, tag.Container))
        self.assertIn('maxcdn.bootstrapcdn.com', self.doc.build())

        elm.innerHTML = '<button is="default-button"></button>'
        self.assertTrue(isinstance(elm.firstChild, bootstrap3.DefaultButton))
        self.assertTrue(isinstance(elm.firstChild, bootstrap3.Button))
        self.assertFalse(isinstance(elm.firstChild, tag.DefaultButton))
        self.assertTrue(isinstance(elm.firstChild, tag.Button))

    def test_tempdir(self):
        doc = Document()
        self.assertIsNotNone(doc.tempdir)
        self.assertTrue(os.path.exists(doc.tempdir))
        self.assertTrue(os.path.isabs(doc.tempdir))
        self.assertTrue(os.path.isdir(doc.tempdir))
        tempdir = doc.tempdir
        testfile = os.path.join(tempdir, 'test_file')
        with open(testfile, 'w') as f:
            f.write('test')
        self.assertTrue(os.path.exists(testfile))
        self.assertTrue(os.path.isfile(testfile))
        with open(testfile) as f:
            self.assertEqual('test', f.read().strip())
        del doc
        import gc
        gc.collect()
        self.assertFalse(os.path.exists(testfile))
        self.assertFalse(os.path.exists(tempdir))


class TestDocumentOptions(TestCase):
    def setUp(self):
        super().setUp()
        options.parse_command_line()

    def test_set_new_document(self):
        old_doc = get_document()
        doc = get_new_document()
        self.assertIsNot(doc, old_doc)
        set_document(doc)
        new_doc = get_document()
        self.assertIsNot(doc, old_doc)
        self.assertIs(doc, new_doc)

    def test_document_autoreload(self):
        doc = get_new_document(autoreload=True)
        html = doc.build()
        self.assertIn('RIMO_AUTORELOAD = true', html)
        self.assertNotIn('RIMO_RELOAD_WAIT', html)

    def test_document_reload_wait(self):
        doc = get_new_document(autoreload=True, reload_wait=1234)
        html = doc.build()
        self.assertIn('RIMO_AUTORELOAD = true', html)
        self.assertIn('RIMO_RELOAD_WAIT = 1234', html)

    def test_document_no_reload_wait_no_reload(self):
        doc = get_new_document(autoreload=False, reload_wait=1234)
        html = doc.build()
        self.assertNotIn('RIMO_AUTORELOAD', html)
        self.assertNotIn('RIMO_RELOAD_WAIT', html)

    def test_document_log_level_str(self):
        doc = get_new_document(log_level='INFO')
        html = doc.build()
        self.assertIn('RIMO_LOG_LEVEL = \'INFO\'', html)

    def test_document_log_level_int(self):
        doc = get_new_document(log_level=10)
        html = doc.build()
        self.assertIn('RIMO_LOG_LEVEL = 10', html)

    def test_document_log_prefix(self):
        doc = get_new_document(log_prefix='TEST')
        html = doc.build()
        self.assertIn('RIMO_LOG_PREFIX = \'TEST\'', html)

    def test_document_log_console(self):
        doc = get_new_document(log_console=True)
        html = doc.build()
        self.assertIn('RIMO_LOG_CONSOLE = true', html)

    def test_document_ws_url(self):
        doc = get_new_document(ws_url='test_ws')
        html = doc.build()
        self.assertIn('RIMO_WS_URL = \'test_ws\'', html)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio

from wdom.document import get_document
from wdom.server import start_server, stop_server
from wdom.themes.default import Div, H1, Container
from wdom.themes.default import PrimaryButton, DangerButton, DefaultButton


class Timer(Container):
    def __init__(self, *args, **kwargs):
        self._running = False
        self._count = 180
        self._interval = 0.01
        self._loop = asyncio.get_event_loop()
        super().__init__(*args, **kwargs)
        self._view = H1(parent=self)
        self.update()
        self._start_btn = PrimaryButton('start', parent=self, id='start_btn')
        self._start_btn.addEventListener('click', self.start)
        self._stop_btn = DangerButton('stop', parent=self, id='stop_btn')
        self._stop_btn.addEventListener('click', self.stop)
        self._reset_btn = DefaultButton('reset', parent=self, id='reset_btn')
        self._reset_btn.addEventListener('click', self.reset)

    def stop(self, event):
        self._running = False

    def start(self, event):
        self._running = True
        self._timer_loop()

    def reset(self, event):
        self._count = 180
        self.update()

    def _timer_loop(self):
        if self._running and self._count > 0:
            self._loop.call_later(self._interval, self._timer_loop)
            self._count -= self._interval
            self.update()

    def set(self, value):
        self._view.textContent = '{0:.2f}'.format(value)

    def update(self):
        self.set(self._count)


def sample_app(**kwargs) -> Div:
    return Timer()


if __name__ == '__main__':
    from wdom.themes import default
    document = get_document()
    document.register_theme(default)
    document.body.prepend(sample_app())
    start_server()
    try:
        asyncio.get_event_loop().run_forever()
    except KeyboardInterrupt:
        pass
    finally:
        stop_server()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'Kube'
project_url = 'https://imperavi.com/kube/'
project_repository = 'https://github.com/imperavi/kube'
license = 'MIT License'
license_url = 'https://imperavi.com/kube/#kube-intro'

css_files = [
    '_static/css/kube.min.css',
]


class TypedButton(Button):
    type_ = None
    outline = False
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.type_:
            self.setAttribute('type', self.type_)
        if self.outline:
            self.setAttribute('outline', True)

DefaultButton = NewTag('DefaultButton', 'button', TypedButton, type_='black',
                       outline=True, is_='default-button')
PrimaryButton = NewTag('PrimaryButton', 'button', TypedButton, type_='primary', is_='primary-button')
SecondaryButton = NewTag('SecondaryButton', 'button', TypedButton, type_='black',
                         outline=True, is_='secondary-button')
SuccessButton = NewTag('SuccessButton', 'button', TypedButton, type_='black', is_='success-button')
InfoButton = NewTag('InfoButton', 'button', TypedButton, type_='primary',
                    outline=True, is_='info-button')
WarningButton = NewTag('WarningButton', 'button', TypedButton, is_='warning-button')
DangerButton = NewTag('DangerButton', 'button', TypedButton, is_='danger-button')
ErrorButton = NewTag('ErrorButton', 'button', TypedButton, is_='error-button')
LinkButton = NewTag('LinkButton', 'button', TypedButton, outline=True, is_='link-button')

Form = NewTag('Form', 'form', Form, class_='forms')
Select = NewTag('Select', 'select', Select, class_='select')

Row = NewTag('Row', 'row', Row)

class Col(Div):
    tag = 'column'
    cols = 0
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAttribute('cols', str(self.cols))

Col1 = NewTag('Col1', 'column', Col, is_='col1')
Col2 = NewTag('Col2', 'column', Col, cols=2, is_='col2')
Col3 = NewTag('Col3', 'column', Col, cols=3, is_='col3')
Col4 = NewTag('Col4', 'column', Col, cols=4, is_='col4')
Col5 = NewTag('Col5', 'column', Col, cols=5, is_='col5')
Col6 = NewTag('Col6', 'column', Col, cols=6, is_='col6')
Col7 = NewTag('Col7', 'column', Col, cols=7, is_='col7')
Col8 = NewTag('Col8', 'column', Col, cols=8, is_='col8')
Col9 = NewTag('Col9', 'column', Col, cols=9, is_='col9')
Col10 = NewTag('Col10', 'column', Col, cols=10, is_='col10')
Col11 = NewTag('Col11', 'column', Col, cols=11, is_='col11')
Col12 = NewTag('Col12', 'column', Col, cols=12, is_='col12')

extended_classes = [
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    Form,
    Select,
    Row,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
    Col12,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
import subprocess
import time
import unittest

from selenium.webdriver.common.utils import free_port

from wdom.testing import TestCase
from wdom.testing import close_webdriver, get_webdriver


@unittest.skipIf(os.environ.get('TRAVIS', False), reason='skip on CI')
def setUpModule():
    pass


def tearDownModule():
    close_webdriver()


class BaseTestCase(object):
    module = ''

    def setUp(self):
        self.wd = get_webdriver()
        self.port = free_port()
        cmd = [sys.executable, '-m', self.module, '--port', str(self.port)]
        self.proc = subprocess.Popen(cmd, env=os.environ,
                                     stdout=subprocess.PIPE,
                                     stderr=subprocess.PIPE,
                                     )
        self.url = 'http://localhost:{}'.format(self.port)
        time.sleep(0.5)
        self.wd.get(self.url)

    def tearDown(self):
        self.proc.terminate()
        super().tearDown()


class TestReverseText(BaseTestCase, TestCase):
    module = 'wdom.examples.rev_text'

    def test_revtext(self):
        text = 'Click!'
        h1 = self.wd.find_element_by_tag_name('h1')
        self.assertEqual(h1.text, text)
        time.sleep(0.5)
        h1.click()
        time.sleep(0.5)
        self.assertEqual(h1.text, text[::-1])
        h1.click()
        time.sleep(0.5)
        self.assertEqual(h1.text, text)


class TestDataBinding(BaseTestCase, TestCase):
    module = 'wdom.examples.data_binding'

    def test_revtext(self):
        h1 = self.wd.find_element_by_tag_name('h1')
        input = self.wd.find_element_by_tag_name('input')
        time.sleep(0.5)
        for k in 'test':
            input.send_keys(k)
        time.sleep(0.5)
        self.assertEqual(h1.text, 'test')


class TestTimer(BaseTestCase, TestCase):
    module = 'wdom.examples.timer'

    def test_timer(self):
        view = self.wd.find_element_by_tag_name('h1')
        start_btn = self.wd.find_element_by_id('start_btn')
        stop_btn = self.wd.find_element_by_id('stop_btn')
        reset_btn = self.wd.find_element_by_id('reset_btn')
        time.sleep(0.5)

        def test_timer():
            self.assertEqual(view.text, '180.00')
            start_btn.click()
            time.sleep(0.5)
            self.assertTrue(float(view.text) < 179.90)
            stop_btn.click()
            t = view.text
            time.sleep(0.5)
            self.assertEqual(view.text, t)

        test_timer()
        reset_btn.click()
        time.sleep(0.5)
        test_timer()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
from os import path
import asyncio
import subprocess
from tempfile import NamedTemporaryFile

from selenium.common.exceptions import NoSuchElementException

from wdom.misc import install_asyncio
from wdom.testing import get_webdriver, free_port, browser_implict_wait
from wdom.testing import TestCase, close_webdriver


def setUpModule():
    install_asyncio()


def tearDownModule():
    close_webdriver()


CURDIR = path.dirname(path.abspath(__file__))

src_base = '''
import sys
import asyncio

from wdom.misc import install_asyncio
from wdom.tag import H1
from wdom.document import get_document
from wdom import server

install_asyncio()
loop = asyncio.get_event_loop()
doc = get_document()
doc.body.appendChild(H1('FIRST', id='h1'))
doc.add_cssfile('testdir/test.css')
server.add_static_path('testdir', '{curdir}/testdir')
server.start_server(loop=loop, check_time=10)
loop.run_forever()
'''.format(curdir=CURDIR)

css_path = path.join(CURDIR, 'testdir/test.css')
src_css = '''
h1 {color: #000000;}
'''
src_css_post = '''
h1 {color: #ff0000;}
'''

_src = src_base.splitlines()
_src.insert(12, 'server.exclude_patterns.append(r\'test.css\')')
src_exclude_css = '\n'.join(_src)
_src = src_base.splitlines()
_src.insert(12, 'server.exclude_patterns.append(r\'testdi*\')')
src_exclude_dir = '\n'.join(_src)


class TestAutoReload(TestCase):
    wait_time = 3 if os.environ.get('TRAVIS') else 1

    @classmethod
    def setUpClass(cls):
        cls.wd = get_webdriver()
        if os.environ.get('TRAVIS', True):
            cls.wd.implicitly_wait(10)

    @classmethod
    def tearDownClass(cls):
        if os.environ.get('TRAVIS', True):
            cls.wd.implicitly_wait(browser_implict_wait)

    def setUp(self):
        super().setUp()
        with open(css_path, 'w') as f:
            f.write(src_css)
        self.port = free_port()
        self.url = 'http://localhost:{}'.format(self.port)
        tmpfile = NamedTemporaryFile(mode='w+', dir=CURDIR, suffix='.py',
                                     delete=False)
        self.tmpfilename = tmpfile.name
        tmpfile.close()
        self.proc = None

    def tearDown(self):
        if self.proc is not None and self.proc.returncode is None:
            self.proc.terminate()
        if path.exists(self.tmpfilename):
            os.remove(self.tmpfilename)
        with open(css_path, 'w') as f:
            f.write(src_css)
        super().tearDown()

    def _base_args(self):
        return [sys.executable, self.tmpfilename, '--port', str(self.port),
                '--logging', 'error']

    def wait(self, t: float = None):
        _t = t or self.wait_time
        for i in range(10):
            asyncio.get_event_loop().run_until_complete(asyncio.sleep(_t/10))

    def wait_short(self, t: float = None):
        asyncio.get_event_loop().run_until_complete(
            asyncio.sleep(t or (self.wait_time / 10)))

    def check_reload(self, args):
        self.proc = subprocess.Popen(args, cwd=CURDIR, env=os.environ)
        self.wait()
        self.wd.get(self.url)
        self.wait()
        try:
            h1 = self.wd.find_element_by_id('h1')
        except NoSuchElementException:
            self.wait()
            h1 = self.wd.find_element_by_id('h1')
        self.assertEqual(h1.text, 'FIRST')

        with open(self.tmpfilename, 'w') as f:
            f.write(src_base.replace('FIRST', 'SECOND'))
        self.wait()
        try:
            h1 = self.wd.find_element_by_id('h1')
        except NoSuchElementException:
            self.wait()
            h1 = self.wd.find_element_by_id('h1')
        h1 = self.wd.find_element_by_id('h1')
        self.assertEqual(h1.text, 'SECOND')

    def test_autoreload(self):
        with open(self.tmpfilename, 'w') as f:
            f.write(src_base)
        self.wait_short()
        args = self._base_args()
        args.append('--autoreload')
        self.check_reload(args)

    def test_autoreload_debug(self):
        with open(self.tmpfilename, 'w') as f:
            f.write(src_base)
        self.wait_short()
        args = self._base_args()
        args.append('--debug')
        self.check_reload(args)

    def check_css_reload(self, args):
        self.proc = subprocess.Popen(args, cwd=CURDIR, env=os.environ)
        self.wait()
        self.wd.get(self.url)
        self.wait()
        try:
            h1 = self.wd.find_element_by_id('h1')
        except NoSuchElementException:
            self.wait()
            h1 = self.wd.find_element_by_id('h1')
        # value_of_css_property return colors as rgba style
        self.assertRegex(h1.value_of_css_property('color'),
                         r'0,\s*0,\s* 0,\s*1\s*')
        with open(css_path, 'w') as f:
            f.write(src_css_post)
        self.wait()
        try:
            h1 = self.wd.find_element_by_id('h1')
        except NoSuchElementException:
            self.wait()
            h1 = self.wd.find_element_by_id('h1')
        self.assertRegex(h1.value_of_css_property('color'),
                         r'255,\s*0,\s* 0,\s*1\s*')

    def test_autoreload_css(self):
        with open(self.tmpfilename, 'w') as f:
            f.write(src_base)
        self.wait_short()
        args = self._base_args()
        args.append('--autoreload')
        self.check_css_reload(args)

    def check_css_noreload(self, args):
        self.proc = subprocess.Popen(args, cwd=CURDIR, env=os.environ)
        self.wait()
        self.wd.get(self.url)
        self.wait()
        try:
            h1 = self.wd.find_element_by_id('h1')
        except NoSuchElementException:
            self.wait()
            h1 = self.wd.find_element_by_id('h1')
        # value_of_css_property return colors as rgba style
        self.assertRegex(h1.value_of_css_property('color'),
                         r'0,\s*0,\s* 0,\s*1\s*')
        with open(css_path, 'w') as f:
            f.write(src_css_post)
        self.wait()
        try:
            h1 = self.wd.find_element_by_id('h1')
        except NoSuchElementException:
            self.wait()
            h1 = self.wd.find_element_by_id('h1')
        self.assertRegex(h1.value_of_css_property('color'),
                         r'0,\s*0,\s* 0,\s*1\s*')

    def test_autoreload_exclude_css(self):
        with open(self.tmpfilename, 'w') as f:
            f.write(src_exclude_css)
        self.wait_short()
        args = self._base_args()
        args.append('--autoreload')
        self.check_css_noreload(args)

    def test_autoreload_exclude_dir(self):
        with open(self.tmpfilename, 'w') as f:
            f.write(src_exclude_dir)
        self.wait_short()
        args = self._base_args()
        args.append('--autoreload')
        self.check_css_noreload(args)

    def test_autoreload_nowatch(self):
        with open(self.tmpfilename, 'w') as f:
            f.write(src_base.replace("/testdir'", "/testdir', no_watch=True"))
        self.wait_short()
        args = self._base_args()
        args.append('--autoreload')
        self.check_css_noreload(args)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'Foundation'
project_url = 'http://foundation.zurb.com/'
project_repository = 'https://github.com/zurb/foundation-sites'
license = 'MIT License'
license_url = 'https://github.com/zurb/foundation-sites/blob/develop/LICENSE'

css_files = [
    '//cdn.jsdelivr.net/foundation/6.2.1/foundation.min.css',
]

js_files = [
    '//ajax.googleapis.com/ajax/libs/jquery/1.11.2/jquery.min.js',
    '//cdn.jsdelivr.net/foundation/6.2.1/foundation.min.js',
]

headers = []

Button = NewTag('Button', bases=Button, class_='button')
DefaultButton = NewTag('DefaultButton', 'button', Button, class_='hollow secondary', is_='default-button')
PrimaryButton = NewTag('PrimaryButton', 'button', Button, is_='primary-button')
SecondaryButton = NewTag('SecondaryButton', 'button', Button, class_='secondary', is_='secondary-button')
SuccessButton = NewTag('SuccessButton', 'button', Button, class_='success', is_='success-button')
InfoButton = NewTag('InfoButton', 'button', Button, class_='success hollow', is_='info-button')
WarningButton = NewTag('WarningButton', 'button', Button, class_='warning', is_='warning-button')
DangerButton = NewTag('DangerButton', 'button', Button, class_='alert', is_='danger-button')
ErrorButton = NewTag('ErrorButton', 'button', Button, class_='alert', is_='error-button')
LinkButton = NewTag('LinkButton', 'button', Button, class_='hollow', is_='link-button')

Row = NewTag('Row', 'div', Row, class_='row')
Col = NewTag('Col', 'div', Col, class_='columns')
Col1 = NewTag('Col1', 'div', Col, class_='small-1 medium-1 large-1', is_='col1')
Col2 = NewTag('Col2', 'div', Col, class_='small-2 medium-2 large-2', is_='col2')
Col3 = NewTag('Col3', 'div', Col, class_='small-3 medium-3 large-3', is_='col3')
Col4 = NewTag('Col4', 'div', Col, class_='small-4 medium-4 large-4', is_='col4')
Col5 = NewTag('Col5', 'div', Col, class_='small-5 medium-5 large-5', is_='col5')
Col6 = NewTag('Col6', 'div', Col, class_='small-6 medium-6 large-6', is_='col6')
Col7 = NewTag('Col7', 'div', Col, class_='small-7 medium-7 large-7', is_='col7')
Col8 = NewTag('Col8', 'div', Col, class_='small-8 medium-8 large-8', is_='col8')
Col9 = NewTag('Col9', 'div', Col, class_='small-9 medium-9 large-9', is_='col9')
Col10 = NewTag('Col10', 'div', Col, class_='small-10 medium-10 large-10', is_='col10')
Col11 = NewTag('Col11', 'div', Col, class_='small-11 medium-11 large-11', is_='col11')
Col12 = NewTag('Col12', 'div', Col, class_='small-12 medium-12 large-12', is_='col12')

extended_classes = [
    Button,
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    Row,
    Col,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
    Col12,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'Material Design Lite (MDL)'
project_url = 'http://www.getmdl.io/'
project_repository = 'https://github.com/google/material-design-lite'
license = 'Apache License 2.0'
license_url = 'https://github.com/google/material-design-lite/blob/master/LICENSE'

css_files = [
    '//fonts.googleapis.com/icon?family=Material+Icons',
    '//code.getmdl.io/1.1.3/material.indigo-pink.min.css',
]
js_files = [
    '//code.getmdl.io/1.1.3/material.min.js',
]
headers = []

Button = NewTag('Button', bases=Button, class_='mdl-button mdl-js-button')
DefaultButton = NewTag('DefaultButton', 'button', Button,
                       class_='mdl-button--raised', is_='default-button')
PrimaryButton = NewTag('PrimaryButton', 'button', DefaultButton,
                       class_='mdl-button--primary', is_='primary-button')
SecondaryButton = NewTag('SecondaryButton', 'button', Button,
                         class_='mdl-button--raised', is_='secondary-button')
SuccessButton = NewTag('SuccessButton', 'button', DefaultButton,
                       class_='mdl-button--accent', is_='success-button')
InfoButton = NewTag('InfoButton', 'button', Button,
                    class_='mdl-button--primary', is_='info-button')
WarningButton = NewTag('WarningButton', 'button', Button, is_='warning-button')
DangerButton = NewTag('DangerButton', 'button', Button, is_='danger-button')
ErrorButton = NewTag('ErrorButton', 'button', Button, is_='error-button')
LinkButton = NewTag('LinkButton', 'button', Button,
                    class_='mdl-button--accent', is_='link-button')

FormGroup = NewTag('FormGroup', 'div', Div, class_='mdl-textfield mdl-js-textfield')
Input = NewTag('Input', 'input', Input, class_='mdl-textfield__input')
Label = NewTag('Label', 'label', Label, class_='mdl-textfield__label')
Textarea = NewTag('Textarea', 'textarea', Textarea, class_='mdl-textfield__input')
Select = NewTag('Select', 'select', Select, class_='mdl-textfield__input')

Table = NewTag('Table', 'table', Table, class_='mdl-data-table mdl-js-data-table')
Th = NewTag('Th', 'th', Th, class_='mdl-data-table__cell--non-numeric')
Td = NewTag('Td', 'td', Td, class_='mdl-data-table__cell--non-numeric')

H1 = NewTag('H1', 'div', H1, class_='mdl-typography--display-4')
H2 = NewTag('H2', 'div', H2, class_='mdl-typography--display-3')
H3 = NewTag('H3', 'div', H3, class_='mdl-typography--display-2')
H4 = NewTag('H4', 'div', H4, class_='mdl-typography--display-1')
H5 = NewTag('H5', 'div', H5, class_='mdl-typography--headline')
H6 = NewTag('H6', 'div', H6, class_='mdl-typography--title')

Row = NewTag('Row', 'div', Row, class_='mdl-grid')
Col = NewTag('Col', 'div', Col, class_='mdl-cell')
Col1 = NewTag('Col1', 'div', Col1, class_='mdl-cell mdl-cell--1-col')
Col2 = NewTag('Col2', 'div', Col2, class_='mdl-cell mdl-cell--2-col')
Col3 = NewTag('Col3', 'div', Col3, class_='mdl-cell mdl-cell--3-col')
Col4 = NewTag('Col4', 'div', Col4, class_='mdl-cell mdl-cell--4-col')
Col5 = NewTag('Col5', 'div', Col5, class_='mdl-cell mdl-cell--5-col')
Col6 = NewTag('Col6', 'div', Col6, class_='mdl-cell mdl-cell--6-col')
Col7 = NewTag('Col7', 'div', Col7, class_='mdl-cell mdl-cell--7-col')
Col8 = NewTag('Col8', 'div', Col8, class_='mdl-cell mdl-cell--8-col')
Col9 = NewTag('Col9', 'div', Col9, class_='mdl-cell mdl-cell--9-col')
Col10 = NewTag('Col10', 'div', Col10, class_='mdl-cell mdl-cell--10-col')
Col11 = NewTag('Col11', 'div', Col11, class_='mdl-cell mdl-cell--11-col')
Col12 = NewTag('Col12', 'div', Col12, class_='mdl-cell mdl-cell--12-col')

extended_classes = [
    Button,
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    FormGroup,
    Input,
    Label,
    Textarea,
    Select,
    Table,
    Th,
    Td,
    H1,
    H2,
    H3,
    H4,
    H5,
    H6,
    Row,
    Col,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
    Col12,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import html
from typing import Union, Tuple

from wdom.interface import Node, NodeList


class Node(Node):
    """Base Abstract Class for Node interface."""
    # DOM Level 1
    nodeType = None
    nodeName = ''
    nodeValue = ''
    attributes = None

    # DOM Level 2
    namespaceURI = ''
    prefix = ''

    # DOM Level 3
    baseURI = ''

    # DOM Level 4
    parentElement = None

    # should escape text contents
    _should_escape_text = False

    def __init__(self, parent=None):
        super().__init__()  # Need to call init in multiple inheritce
        self._children = list()
        self._parent = None
        if parent:
            parent.appendChild(self)

    def __bool__(self) -> bool:
        """Return always True."""
        return True

    def __len__(self) -> int:
        """Return number of child nodes."""
        return self.length

    def __contains__(self, other: Node) -> bool:
        return other in self._children

    def __copy__(self) -> Node:
        clone = type(self)()
        return clone

    def __deepcopy__(self, memo=None) -> Node:
        clone = self.__copy__()
        for child in self.childNodes:
            clone.appendChild(child.__deepcopy__(memo))
        return clone

    # DOM Level 1
    @property
    def length(self) -> int:
        """Return number of child nodes."""
        return len(self.childNodes)

    @property
    def parentNode(self) -> Node:
        """Return parent node.

        If this node does not have a parent, return ``None``.
        """
        return self._parent

    @property
    def childNodes(self) -> NodeList:
        """Return child nodes of this node.

        Returned object is an instance of NodeList, which is a list like object
        but not support any modification. NodeList is a **live object**, which
        means that changes on this node is reflected to the object.
        """
        return NodeList(self._children)

    @property
    def firstChild(self) -> Node:
        """Return the first child node.

        If this node does not have any child, return ``None``.
        """
        if self.hasChildNodes():
            return self.childNodes[0]
        else:
            return None

    @property
    def lastChild(self) -> Node:
        """Return the last child node.

        If this node does not have any child, return ``None``.
        """
        if self.hasChildNodes():
            return self.childNodes[-1]
        else:
            return None

    @property
    def previousSibling(self) -> Node:
        """Return the previous sibling of this node.

        If there is no previous sibling, return ``None``.
        """
        parent = self.parentNode
        if parent is None:
            return None
        return parent.childNodes.item(parent.childNodes.index(self) - 1)

    @property
    def nextSibling(self) -> Node:
        """Return the next sibling of this node.

        If there is no next sibling, return ``None``.
        """
        parent = self.parentNode
        if parent is None:
            return None
        return parent.childNodes.item(parent.childNodes.index(self) + 1)

    # DOM Level 2
    @property
    def ownerDocument(self) -> Node:
        """Return the owner document of this node.

        Owner document is an ancestor document node of this node. If this node
        (or node tree including this node) is not appended to any document
        node, this property returns ``None``.
        """
        if self.nodeType == Node.DOCUMENT_NODE:
            return self
        elif self.parentNode:
            return self.parentNode.ownerDocument
        else:
            return None

    # Methods
    def _append_document_fragment(self, node: Node) -> Node:
        for c in tuple(node.childNodes):
            self._append_child(c)
        return node

    def _append_element(self, node: Node) -> Node:
        if node.parentNode:
            node.parentNode.removeChild(node)
        self._children.append(node)
        node._parent = self
        return node

    def _append_child(self, node: Node) -> Node:
        if not isinstance(node, Node):
            raise TypeError(
                'appndChild() only accepts a Node instance, but get {}. '
                'If you want to add string or mupltiple nodes once, '
                'use append() method instead.'.format(type(node)))
        if node.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
            return self._append_document_fragment(node)
        else:
            return self._append_element(node)

    def appendChild(self, node: Node) -> Node:
        """Append the node at the last of this child nodes."""
        return self._append_child(node)

    def index(self, node: Node):
        """Return index of the node.

        If the node is not a child of this node, raise ``ValueError``.
        """
        if node in self.childNodes:
            return self.childNodes.index(node)
        elif isinstance(node, Text):
            for i, n in enumerate(self.childNodes):
                # should consider multiple match?
                if isinstance(n, Text) and n.data == node:
                    return i
        raise ValueError('node is not in this node')

    def _insert_document_fragment_before(self, node: Node, ref_node: Node
                                         ) -> Node:
        for c in tuple(node.childNodes):
            self._insert_before(c, ref_node)
        return node

    def _insert_element_before(self, node: Node, ref_node: Node) -> Node:
        if node.parentNode:
            node.parentNode.removeChild(node)
        self._children.insert(self.index(ref_node), node)
        node._parent = self
        return node

    def _insert_before(self, node: Node, ref_node: Node) -> Node:
        if not isinstance(node, Node):
            raise TypeError(
                'insertBefore() only accepts a Node instance, but get {}.'
                'If you want to insert string or mupltiple nodes, '
                'use ref_node.before() instead.'.format(type(node)))
        if node.nodeType == Node.DOCUMENT_FRAGMENT_NODE:
            return self._insert_document_fragment_before(node, ref_node)
        else:
            return self._insert_element_before(node, ref_node)

    def insertBefore(self, node: Node, ref_node: Node) -> Node:
        """Insert a node just before the reference node."""
        return self._insert_before(node, ref_node)

    def hasChildNodes(self) -> bool:
        """Return True if this node has child nodes, otherwise return False."""
        return bool(self.childNodes)

    def _remove_child(self, node: Node) -> Node:
        if node not in self._children:
            raise ValueError('node to be removed is not a child of this node.')
        self._children.remove(node)
        node._parent = None
        return node

    def removeChild(self, node: Node) -> Node:
        """Remove a node from this node.

        If node is not a child of this node, raise ``ValueError``.
        """
        return self._remove_child(node)

    def _replace_child(self, new_child: Node, old_child: Node) -> Node:
        self._insert_before(new_child, old_child)
        return self._remove_child(old_child)

    def replaceChild(self, new_child: Node, old_child: Node) -> Node:
        """Replace an old child with new child."""
        return self._replace_child(new_child, old_child)

    def hasAttributes(self) -> bool:
        """Return True if this node has attributes."""
        return bool(self.attributes)

    def cloneNode(self, deep: bool=False) -> Node:
        """Return new copy of this node.

        If optional argument ``deep`` is specified and is True, new node has
        clones of child nodes of this node (if presents).
        """
        if deep:
            return self.__deepcopy__()
        else:
            return self.__copy__()

    def _empty(self):
        for child in tuple(self._children):
            self._remove_child(child)

    def empty(self):
        """[Not Standard] Remove all child nodes from this node.

        This is equivalent to ``node.textContent = ''``.
        """
        self._empty()

    def _get_text_content(self) -> str:
        return ''.join(child.textContent for child in self.childNodes)

    def _set_text_content(self, value: str):
        self._empty()
        if value:
            self._append_child(Text(value))

    @property
    def textContent(self) -> str:
        """Return text contents of this node and child nodes."""
        return self._get_text_content()

    @textContent.setter
    def textContent(self, value: str):
        """Remove all child nodes and set new text."""
        self._set_text_content(value)


def _ensure_node(node: Union[str, 'ChildNode']) -> 'ChildNode':
    if isinstance(node, str):
        return Text(node)
    elif isinstance(node, Node):
        return node
    else:
        raise TypeError('Invalid type to append: {}'.format(node))


def _to_node_list(nodes: Tuple[str, 'ChildNode']) -> Node:
    if len(nodes) == 1:
        return _ensure_node(nodes[0])
    else:
        df = DocumentFragment()
        for n in nodes:
            df.appendChild(_ensure_node(n))
        return df


class ParentNode:
    """[DOM Level 4] Mixin class for Document, DocumentFragment, and Element.
    """
    @property
    def children(self):
        """Currently this is not a live object"""
        return NodeList([
            e for e in self.childNodes if e.nodeType == Node.ELEMENT_NODE
        ])

    @property
    def firstElementChild(self):
        """First Element child node.

        If this node has no element child, return None.
        """
        for child in self.childNodes:
            if child.nodeType == Node.ELEMENT_NODE:
                return child

    @property
    def lastElementChild(self):
        """Last Element child node.

        If this node has no element child, return None.
        """
        for child in reversed(self.childNodes):
            if child.nodeType == Node.ELEMENT_NODE:
                return child

    def prepend(self, *nodes: Tuple['ChildNode', str]):
        """Insert new nodes before first child node."""
        node = _to_node_list(nodes)
        if self.firstChild:
            self.insertBefore(node, self.firstChild)
        else:
            self.appendChild(node)

    def append(self, *nodes: Tuple['ChildNode', str]):
        """Append new nodes after last child node."""
        node = _to_node_list(nodes)
        self.appendChild(node)

    @property
    def query(self, relativeSelectors):
        raise NotImplementedError

    @property
    def queryAll(self, relativeSelectors):
        raise NotImplementedError

    @property
    def querySelector(self, selectors):
        raise NotImplementedError

    @property
    def querySelectorAll(self, selectors):
        raise NotImplementedError


class NonDocumentTypeChildNode:
    @property
    def previousElementSibling(self) -> Node:
        """Previous Element Node.

        If this node has no previous element node, return None.
        """
        if self.parentNode is None:
            return None
        siblings = self.parentNode.childNodes
        for i in range(siblings.index(self), 0, -1):
            n = siblings[i-1]
            if n.nodeType == Node.ELEMENT_NODE:
                return n

    @property
    def nextElementSibling(self) -> Node:
        """Next Element Node.

        If this node has no next element node, return None.
        """
        if self.parentNode is None:
            return None
        siblings = self.parentNode.childNodes
        for i in range(siblings.index(self) + 1, len(siblings)):
            n = siblings[i]
            if n.nodeType == Node.ELEMENT_NODE:
                return n


class ChildNode:
    """[DOM Level 4] Mixin class for DocumentType, Element, and CharacterData
    (Text, RawHTML, Comment).
    """

    def before(self, *nodes: Tuple['ChildNode', str]) -> None:
        """Insert nodes before this node.

        If nodes contains ``str``, it will be converted to Text node.
        """
        if self.parentNode:
            node = _to_node_list(nodes)
            self.parentNode.insertBefore(node, self)

    def after(self, *nodes: Tuple['ChildNode', str]) -> None:
        """Append nodes after this node.

        If nodes contains ``str``, it will be converted to Text node.
        """
        if self.parentNode:
            node = _to_node_list(nodes)
            _next_node = self.nextSibling
            if _next_node is None:
                self.parentNode.appendChild(node)
            else:
                self.parentNode.insertBefore(node, _next_node)

    def replaceWith(self, *nodes: Tuple['ChildNode', str]) -> None:
        """Replace this node with nodes.

        If nodes contains ``str``, it will be converted to Text node.
        """
        if self.parentNode:
            node = _to_node_list(nodes)
            self.parentNode.replaceChild(node, self)

    def _remove(self):
        if self.parentNode:
            self.parentNode.removeChild(self)

    def remove(self) -> None:
        """Remove this node from the parent node."""
        self._remove()


class CharacterData(Node, ChildNode, NonDocumentTypeChildNode):
    # DOM Level 1
    firstChild = None
    lastChild = None
    specified = False

    def __init__(self, text: str='', parent=None):
        super().__init__(parent=parent)
        self.data = text

    def __copy__(self):
        clone = type(self)(self.data)
        return clone

    @property
    def html(self) -> str:
        return self.textContent

    def _get_text_content(self) -> str:
        return self.data

    def _set_text_content(self, value: str):
        self.data = value

    def __len__(self) -> int:
        return len(self.data)

    @property
    def length(self) -> int:
        return len(self)

    def _append_data(self, string: str):
        self.data += string

    def appendData(self, string: str):
        self._append_data(string)

    def _insert_data(self, offset: int, string: str):
        self.data = ''.join((self.data[:offset], string, self.data[offset:]))

    def insertData(self, offset: int, string: str):
        self._insert_data(offset, string)

    def _delete_data(self, offset: int, count: int):
        self.data = ''.join((self.data[:offset], self.data[offset+count:]))

    def deleteData(self, offset: int, count: int):
        self._delete_data(offset, count)

    def _replace_data(self, offset: int, count: int, string: str):
        self.data = ''.join((
            self.data[:offset], string, self.data[offset+count:]))

    def replaceData(self, offset: int, count: int, string: str):
        self._replace_data(offset, count, string)

    @property
    def childNodes(self) -> NodeList:
        return NodeList([])

    # Methods
    def appendChild(self, node) -> None:
        raise NotImplementedError('This node does not support this method.')

    def insertBefore(self, node, ref_node) -> None:
        raise NotImplementedError('This node does not support this method.')

    def hasChildNodes(self) -> bool:
        return False

    def removeChild(self, node) -> None:
        raise NotImplementedError('This node does not support this method.')

    def replaceChild(self, old_child: Node, new_child: Node) -> None:
        raise NotImplementedError('This node does not support this method.')

    def hasAttributes(self) -> bool:
        return False


class Text(CharacterData):
    nodeType = Node.TEXT_NODE
    nodeName = '#text'

    @property
    def html(self) -> str:
        if self.parentNode and self.parentNode._should_escape_text:
            return html.escape(self.data)
        else:
            return self.data


class RawHtml(Text):
    """Very similar to ``Text`` class, but contents are not escaped.

    Used for inner contents of ``<script>`` element or ``<style>`` element.
    """
    @property
    def html(self) -> str:
        return self.data


class Comment(CharacterData):
    nodeType = Node.COMMENT_NODE
    nodeName = '#comment'

    @property
    def html(self) -> str:
        return ''.join(('<!--', self.data, '-->'))


class DocumentType(Node, NonDocumentTypeChildNode):
    nodeType = Node.DOCUMENT_TYPE_NODE
    nodeValue = None
    textContent = None
    _should_escape_text = True

    @property
    def nodeName(self) -> str:
        return self.name

    def __init__(self, type='html', parent=None):
        super().__init__(parent=parent)
        self._type = type

    @property
    def name(self) -> str:
        return self._type

    @name.setter
    def name(self, name: str):
        self._type = name

    @property
    def html(self) -> str:
        return '<!DOCTYPE {}>'.format(self.name)


class DocumentFragment(Node, ParentNode):
    nodeType = Node.DOCUMENT_FRAGMENT_NODE
    nodeName = '#document-fragment'
    parentNode = None
    previousSibling = None
    nextSibling = None
    _should_escape_text = True

    @property
    def html(self) -> str:
        return ''.join(child.html for child in self.childNodes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from os import path


# if freezed by cx_freeze, _static and _template dirs are copied to lib dir.
if getattr(sys, 'frozen', False):
    root_dir = path.join(path.dirname(sys.executable), 'lib')
else:
    root_dir = path.dirname(path.abspath(__file__))

static_dir = path.join(root_dir, '_static')
template_dir = path.join(root_dir, '_template')

'''
Please include these directories when freeze your app by cx_freeze.

Example:

    from cx_Freeze import setup
    form wdom.misc import include_dirs
    setup(..., options = {'build_exe': {'include_files': include_dirs}}, ...)

'''
include_dirs = [static_dir, template_dir]


def install_asyncio():
    from tornado.ioloop import IOLoop
    from tornado.platform.asyncio import AsyncIOMainLoop
    '''Ensure that asyncio's io-loop is installed to tornado.'''
    if not IOLoop.initialized():
        AsyncIOMainLoop().install()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
import logging
import socket
from typing import Optional

from tornado import web
from tornado import websocket
from tornado.httpserver import HTTPServer

from wdom.options import config
from wdom.misc import install_asyncio
from wdom.server.handler import on_websocket_message

logger = logging.getLogger(__name__)
install_asyncio()
connections = []


def is_connected():
    """Check if tornado web server has a client connection."""
    return any(connections)


class MainHandler(web.RequestHandler):
    """Main handler to serve document of the application."""
    def get(self):
        """Return whole html representation of the root document."""
        from wdom.document import get_document
        logger.info('connected')
        self.write(get_document().build())


class WSHandler(websocket.WebSocketHandler):
    """Handler class of web socket connection."""
    def open(self):
        """Called when connection open."""
        logger.info('WS OPEN')
        connections.append(self)

    def on_message(self, message):
        """Called when get message from client."""
        on_websocket_message(message)

    @asyncio.coroutine
    def terminate(self):
        """Terminate server if no more connection exists."""
        yield from asyncio.sleep(config.shutdown_wait)
        # stop server and close loop if no more connection exists
        if not is_connected():
            stop_server(self.application.server)
            self.application.server.io_loop.stop()

    def on_close(self):
        """Called when connection closed."""
        logger.info('RootWS CLOSED')
        if self in connections:
            # Remove this connection from connection-list
            connections.remove(self)
        # close if auto_shutdown is enabled and there is no more connection
        if config.auto_shutdown and not is_connected():
            asyncio.ensure_future(self.terminate())


class Application(web.Application):
    """Application for a tornado web server.

    This class is based on tornado.web.Application, but including some utility
    methods to make it easy to set up app.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(
            [(r'/', MainHandler), (r'/rimo_ws', WSHandler)],
            *args,
            debug=config.debug,
            autoreload=False,
            static_hash_cashe=False,
            compiled_template_cashe=False,
            **kwargs
        )

    def log_request(self, handler):
        """Handle access log."""
        if 'log_function' in self.settings:
            self.settings['log_function'](handler)
            return
        status = handler.get_status()
        if status < 400:
            log_method = logger.info
        elif status < 500:
            log_method = logger.warning
        else:
            log_method = logger.error
        request_time = 1000.0 * handler.request.request_time()
        if request_time > 10:
            logger.warning('%d %s %.2fms',
                           status, handler._request_summary(), request_time)
        else:
            log_method('%d %s', status, handler._request_summary())

    def add_static_path(self, prefix: str, path: str):
        """Add path to serve static files.

        ``prefix`` is used for url prefix to serve static files and ``path`` is
        a path to the static file directory. ``prefix = '/_static'`` is
        reserved for the server, so do not use it for your app.
        """
        pattern = prefix
        if not pattern.startswith('/'):
            pattern = '/' + pattern
        if not pattern.endswith('/(.*)'):
            pattern = pattern + '/(.*)'
        spec = web.URLSpec(pattern, web.StaticFileHandler, dict(path=path))
        # Need some check
        handlers = self.handlers[0][1]
        handlers.append(spec)

    def add_favicon_path(self, path: str):
        """Add path to serve favicon file.

        ``path`` should be a directory, which contains favicon file
        (``favicon.ico``) for your app.
        """
        spec = web.URLSpec(
            '/(favicon.ico)',
            web.StaticFileHandler,
            dict(path=path)
        )
        # Need some check
        handlers = self.handlers[0][1]
        handlers.append(spec)


main_application = Application()


def get_app(*args, **kwargs) -> Application:
    """Return Application object to serve ``document``."""
    return main_application


def set_application(app: Application):
    """Set application as a root application for the server."""
    global main_application
    main_application = app


def start_server(app: web.Application = None, port: int = None,
                 browser: str = None, address: str = None,
                 check_time: Optional[int] = 500,
                 **kwargs) -> HTTPServer:
    """Start server with ``app`` on ``localhost:port``.

    If port is not specified, use command line option of ``--port``.

    If ``browser`` is not specified, do not open the page. When ``browser`` is
    specified, open the page with the specified browser. The specified browser
    name is not registered in ``webbrowser`` module, or, for example it is just
    ``True``, use system's default browser to open the page.
    """
    app = app or get_app()
    port = port if port is not None else config.port
    address = address if address is not None else config.address

    server = app.listen(port, address=address)
    app.server = server
    server.address = address
    for sock in server._sockets.values():
        if sock.family == socket.AF_INET:
            server.port = sock.getsockname()[1]
            break
    return server


def stop_server(server: HTTPServer):
    """Terminate given server."""
    server.stop()
    logger.info('Server terminated')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'Pure'
project_url = 'http://purecss.io/'
project_repository = 'https://github.com/yahoo/pure/'
license = 'BSD License'
license_url = 'https://github.com/yahoo/pure/blob/master/LICENSE.md'

css_files = [
    '//yui.yahooapis.com/pure/0.6.0/pure-min.css',
    '_static/css/pure-extra.css',
]

Button = NewTag('Button', bases=Button, class_='pure-button')
DefaultButton = NewTag('DefaultButton', 'button', Button, is_='default-button')
PrimaryButton = NewTag('PrimaryButton', 'button', Button, class_='pure-button-primary', is_='primary-button')
SecondaryButton = NewTag('SecondaryButton', 'button', Button, class_='pure-button-secondary', is_='secondary-button')
SuccessButton = NewTag('SuccessButton', 'button', Button, class_='pure-button-success', is_='success-button')
InfoButton = NewTag('InfoButton', 'button', Button, class_='pure-button-secondary', is_='info-button')
WarningButton = NewTag('WarningButton', 'button', Button, class_='pure-button-warning', is_='warning-button')
DangerButton = NewTag('DangerButton', 'button', Button, class_='pure-button-error', is_='danger-button')
ErrorButton = NewTag('ErrorButton', 'button', Button, class_='pure-button-error', is_='error-button')
LinkButton = NewTag('LinkButton', 'button', Button, class_='pure-button-link', is_='link-button')

Form = NewTag('Form', 'form', Form, class_='pure-form')

Table = NewTag('Table', 'table', Table, class_='pure-table')

Row = NewTag('Row', 'div', Row, class_='pure-g')
Col1 = NewTag('Col1', 'div', Col1, class_='pure-u-2-24')
Col2 = NewTag('Col2', 'div', Col2, class_='pure-u-4-24')
Col3 = NewTag('Col3', 'div', Col3, class_='pure-u-6-24')
Col4 = NewTag('Col4', 'div', Col4, class_='pure-u-8-24')
Col5 = NewTag('Col5', 'div', Col5, class_='pure-u-10-24')
Col6 = NewTag('Col6', 'div', Col6, class_='pure-u-12-24')
Col7 = NewTag('Col7', 'div', Col7, class_='pure-u-14-24')
Col8 = NewTag('Col8', 'div', Col8, class_='pure-u-16-24')
Col9 = NewTag('Col9', 'div', Col9, class_='pure-u-18-24')
Col10 = NewTag('Col10', 'div', Col10, class_='pure-u-20-24')
Col11 = NewTag('Col11', 'div', Col11, class_='pure-u-22-24')
Col12 = NewTag('Col12', 'div', Col12, class_='pure-u-24-24')


extended_classes = [
    Button,
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    Form,
    Table,
    Row,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
    Col12,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import os
from os import path
import re
import json
import time
import subprocess
import asyncio
import tempfile

from selenium.webdriver.common.utils import free_port
from syncer import sync

from wdom.misc import install_asyncio
from wdom.document import get_document
from wdom import server
from wdom.testing import TestCase, HTTPTestCase

curdir = path.dirname(__file__)
root = path.dirname(path.dirname(curdir))
script = '''
import asyncio
import atexit
from wdom import document, server
doc = document.get_document()
with open(doc.tempdir + '/a.html', 'w') as f:
    f.write(doc.tempdir)
server.start_server()
atexit.register(server.stop_server)
try:
    asyncio.get_event_loop().run_forever()
except:
    server.stop_server()
'''


def setUpModule():
    install_asyncio()


class TestServerTypeSet(TestCase):
    def test_server_module(self):
        from wdom.server import _tornado
        server.set_server_type('tornado')
        self.assertTrue(isinstance(server.get_app(), _tornado.Application))
        try:
            from wdom.server import _aiohttp
            server.set_server_type('aiohttp')
            self.assertTrue(isinstance(server.get_app(), _aiohttp.Application))
        except ImportError:
            pass
        # here server type is the same as original one

    def test_invalid_server_type(self):
        with self.assertRaises(ValueError):
            server.set_server_type('a')


class TestServerBase(HTTPTestCase):
    cmd = []

    def setUp(self):
        super().setUp()
        self.port = free_port()
        env = os.environ.copy()
        env['PYTHONPATH'] = root
        _ = tempfile.NamedTemporaryFile(mode='w+', suffix='.py', delete=False)
        with _ as f:
            self.tmp = f.name
            f.write(script)
        cmd = [sys.executable, self.tmp, '--port', str(self.port)] + self.cmd
        self.url = 'http://localhost:{}'.format(self.port)
        self.ws_url = 'ws://localhost:{}/rimo_ws'.format(self.port)
        self.proc = subprocess.Popen(
            cmd, cwd=curdir, env=env,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            universal_newlines=True,
        )
        sync(self.wait(times=10))

    def tearDown(self):
        if os.path.exists(self.tmp):
            os.remove(self.tmp)
        if self.proc.returncode is None:
            self.proc.terminate()
        sync(self.wait(times=10))
        super().tearDown()


class TestAutoShutdown(TestServerBase):
    cmd = ['--auto-shutdown', '--shutdown-wait', '0.1']

    @sync
    @asyncio.coroutine
    def test_autoshutdown(self):
        yield from self.wait()
        ws = yield from self.ws_connect(self.ws_url)
        ws.close()
        yield from self.wait(timeout=0.1, times=5)
        self.assertIsNotNone(self.proc.poll())

    @sync
    @asyncio.coroutine
    def test_reload(self):
        yield from self.wait()
        ws = yield from self.ws_connect(self.ws_url)
        ws.close()
        ws = yield from self.ws_connect(self.ws_url)
        yield from self.wait(timeout=0.1, times=5)
        self.assertIsNone(self.proc.poll())

    @sync
    @asyncio.coroutine
    def test_multi_connection(self):
        yield from self.wait()
        ws1 = yield from self.ws_connect(self.ws_url)
        ws2 = yield from self.ws_connect(self.ws_url)
        ws1.close()
        yield from self.wait(timeout=0.1, times=5)
        self.assertIsNone(self.proc.poll())
        ws2.close()
        yield from self.wait(timeout=0.1, times=5)
        self.assertIsNotNone(self.proc.poll())

    @sync
    @asyncio.coroutine
    def test_tempdir_cleanup(self):
        yield from self.wait()
        ws = yield from self.ws_connect(self.ws_url)
        resp = yield from self.fetch(self.url + '/tmp/a.html')
        content = resp.body
        self.assertTrue(path.exists(content.strip()))
        self.assertTrue(path.isdir(content.strip()))
        ws.close()
        yield from self.wait(timeout=0.1, times=5)
        self.assertIsNotNone(self.proc.poll())
        self.assertFalse(path.exists(content.strip()))
        self.assertFalse(path.isdir(content.strip()))


class TestOpenBrowser(TestServerBase):
    cmd = ['--debug', '--open-browser', '--browser', 'firefox']

    def test_open_browser(self):
        time.sleep(0.5)
        self.assertIn('Start server on', self.proc.stdout.readline())
        self.assertIn('connected', self.proc.stdout.readline())


class TestOpenBrowserFreePort(TestServerBase):
    cmd = ['--port', '0', '--open-browser', '--browser', 'firefox']

    def test_open_browser_free_port(self):
        time.sleep(0.5)
        self.assertIn('Start server on', self.proc.stdout.readline())
        self.assertIn('connected', self.proc.stdout.readline())


class TestMainHandlerBlank(HTTPTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.document = get_document()
        self.start()

    @sync
    @asyncio.coroutine
    def test_blank_mainpage(self) -> None:
        with self.assertLogs('wdom', 'INFO'):
            res = yield from self.fetch(self.url)
        self.assertEqual(res.code, 200)
        _re = re.compile(
            '<!DOCTYPE html>\s*<html rimo_id="\d+">\s*<head rimo_id="\d+">'
            '.*<meta .*<title rimo_id="\d+">\s*W-DOM\s*</title>.*'
            '</head>\s*<body.*>.*<script.*>.*</script>.*'
            '</body>\s*</html>',
            re.S)
        self.assertIsNotNone(_re.match(res.text))


class TestMainHandler(HTTPTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.document = get_document()
        self.document.body.prepend('testing')
        self.start()

    @sync
    @asyncio.coroutine
    def test_blank_mainpage(self) -> None:
        with self.assertLogs('wdom', 'INFO'):
            res = yield from self.fetch(self.url)
        self.assertEqual(res.code, 200)
        self.assertIn('testing', res.text)


class TestStaticFileHandler(HTTPTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.document = get_document()
        self.start()

    @sync
    @asyncio.coroutine
    def test_static_file(self) -> None:
        with self.assertLogs('wdom.server', 'INFO'):
            res = yield from self.fetch(self.url + '/_static/js/rimo/rimo.js')
        self.assertEqual(res.code, 200)
        self.assertIn('rimo', res.text)
        self.assertIn('rimo.log', res.text)

    @sync
    @asyncio.coroutine
    def test_add_static_path(self) -> None:
        from os import path
        server.add_static_path('a', path.abspath(path.dirname(__file__)))
        with self.assertLogs('wdom.server', 'INFO'):
            res = yield from self.fetch(self.url + '/a/' + __file__)
        self.assertEqual(res.code, 200)
        self.assertIn('this text', res.text)

    @sync
    @asyncio.coroutine
    def test_tempdir(self):
        from os import path
        self.assertTrue(path.exists(self.document.tempdir))
        with self.assertLogs('wdom.server', 'INFO'):
            res = yield from self.fetch(self.url + '/tmp/a.html')
        self.assertEqual(res.code, 404)
        self.assertIn('404', res.text)
        self.assertIn('Not Found', res.text)
        tmp = path.join(self.document.tempdir, 'a.html')
        with open(tmp, 'w') as f:
            f.write('test')
        self.assertTrue(path.exists(tmp))
        with self.assertLogs('wdom.server', 'INFO'):
            res = yield from self.fetch(self.url + '/tmp/a.html')
        self.assertEqual(res.code, 200)
        self.assertEqual('test', res.text)

    @sync
    @asyncio.coroutine
    def test_tempfile(self):
        doc = get_document()
        self.assertTrue(path.exists(doc.tempdir))
        tmp = path.join(doc.tempdir, 'a.html')
        self.assertFalse(path.exists(tmp))
        with open(tmp, 'w') as f:
            f.write('test')
        self.assertTrue(path.exists(tmp))
        response = yield from self.fetch(self.url + '/tmp/a.html')
        self.assertEqual(response.code, 200)
        self.assertEqual(response.text, 'test')

    @sync
    @asyncio.coroutine
    def test_tempfile_404(self):
        response = yield from self.fetch(self.url + '/tmp/b.html')
        self.assertEqual(response.code, 404)
        response = yield from self.fetch(self.url + '/tmp/a.html')
        self.assertEqual(response.code, 404)


class TestRootWSHandler(HTTPTestCase):
    def setUp(self) -> None:
        super().setUp()
        self.start()
        sync(self.wait())
        self.ws_url = 'ws://localhost:{}/rimo_ws'.format(self.port)
        self.ws = sync(self.ws_connect(self.ws_url))

    @sync
    @asyncio.coroutine
    def test_ws_connection(self) -> None:
        with self.assertLogs('wdom.server', 'INFO'):
            _ = yield from self.ws_connect(self.ws_url)
            del _
            yield from self.wait()

    @sync
    @asyncio.coroutine
    def test_logging_error(self) -> None:
        with self.assertLogs('wdom.server', 'INFO'):
            self.ws.write_message(json.dumps(
                dict(type='log', level='error', message='test')
            ))
            yield from self.wait()

    @sync
    @asyncio.coroutine
    def test_logging_warn(self) -> None:
        with self.assertLogs('wdom.server', 'INFO'):
            self.ws.write_message(json.dumps(
                dict(type='log', level='warn', message='test')
            ))
            yield from self.wait()

    @sync
    @asyncio.coroutine
    def test_logging_info(self) -> None:
        with self.assertLogs('wdom.server', 'INFO'):
            self.ws.write_message(json.dumps(
                dict(type='log', level='info', message='test')
            ))
            yield from self.wait()

    @sync
    @asyncio.coroutine
    def test_logging_debug(self) -> None:
        with self.assertLogs('wdom.server', 'DEBUG'):
            self.ws.write_message(json.dumps(
                dict(type='log', level='debug', message='test')
            ))
            yield from self.wait()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'Schema'
project_url = 'http://danmalarkey.github.io/schema/'
project_repository = 'https://github.com/danmalarkey/schema'
license = 'MIT License & CC BY 3.0 (EasyDropdown)'
license_url = 'https://github.com/danmalarkey/schema#license'

css_files = [
    '_static/css/schema.min.css',
]

Button = NewTag('Button', bases=Button, class_='btn')
DefaultButton = NewTag('DefaultButton', 'button', Button, class_='btn-default', is_='default-button')
PrimaryButton = NewTag('PrimaryButton', 'button', Button, class_='btn-primary', is_='primary-button')
SecondaryButton = NewTag('SecondaryButton', 'button', Button, class_='btn-default', is_='secondary-button')
SuccessButton = NewTag('SuccessButton', 'button', Button, class_='btn-success', is_='success-button')
InfoButton = NewTag('InfoButton', 'button', Button, class_='btn-info', is_='info-button')
WarningButton = NewTag('WarningButton', 'button', Button, class_='btn-warning', is_='warning-button')
DangerButton = NewTag('DangerButton', 'button', Button, class_='btn-danger', is_='danger-button')
ErrorButton = NewTag('ErrorButton', 'button', Button, class_='btn-danger', is_='error-button')
LinkButton = NewTag('LinkButton', 'button', Button, class_='btn-link', is_='link-button')

FormGroup = NewTag('FormGroup', 'div', FormGroup, class_='form-group')
Input = NewTag('Input', 'input', Input, class_='form-element')
TextInput = NewTag('TextInput', 'input', TextInput, class_='form-element')
Textarea = NewTag('Textarea', 'textarea', Textarea, class_='form-element')
Select = NewTag('Select', 'select', Select, class_='form-element')

# Lists
# Ul = NewTag('Ul', 'ul', Ul, class_='list-group')
# Li = NewTag('Li', 'li', Li, class_='list-group-element')

Table = NewTag('Table', 'table', Table, class_='table')

Container = NewTag('Container', 'div', Container, class_='container')
Wrapper = NewTag('Wrapper', 'div', Wrapper, class_='container')
Row = NewTag('Row', 'div', Row, class_='row-fluid')

Col1 = NewTag('Col1', 'div', Col1, class_='col1')
Col2 = NewTag('Col2', 'div', Col2, class_='col2')
Col3 = NewTag('Col3', 'div', Col3, class_='col3')
Col4 = NewTag('Col4', 'div', Col4, class_='col4')
Col5 = NewTag('Col5', 'div', Col5, class_='col5')
Col6 = NewTag('Col6', 'div', Col6, class_='col6')
Col7 = NewTag('Col7', 'div', Col7, class_='col7')
Col8 = NewTag('Col8', 'div', Col8, class_='col8')
Col9 = NewTag('Col9', 'div', Col9, class_='col9')
Col10 = NewTag('Col10', 'div', Col10, class_='col10')
Col11 = NewTag('Col11', 'div', Col11, class_='col11')
Col12 = NewTag('Col12', 'div', Col12, class_='col12')

extended_classes = [
    Button,
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    FormGroup,
    Input,
    TextInput,
    Textarea,
    Select,
    Table,
    Container,
    Wrapper,
    Row,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
    Col12,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import logging
import importlib
from typing import Optional

from tornado import autoreload

from wdom.misc import static_dir, install_asyncio
from wdom.options import config
from wdom.server.base import exclude_patterns, open_browser, watch_dir

logger = logging.getLogger(__name__)

try:
    import aiohttp
    v = aiohttp.__version__.split('.')
    if int(v[0]) > 0 or int(v[1]) >= 21:
        from wdom.server import _aiohttp as module
    else:
        logger.warning(
            'wdom requires aiohttp >= 0.21.0, but {} is installed. '
            'please update it by `pip install -U aiohttp.'.format(
                aiohttp.__version__)
        )
        raise ImportError
except ImportError:
    from wdom.server import _tornado as module

__all__ = ('get_app', 'start_server', 'stop_server', 'exclude_patterns')
_server = None


def is_connected():
    """Check if the current server has a client connection."""
    return module.is_connected()


def send_message(msg: str):
    """Send message to all client connections."""
    for conn in module.connections:
        conn.write_message(msg)


def add_static_path(prefix, path, no_watch: bool = False):
    """Add directory to serve static files.

    First argument ``prefix`` is a URL prefix for the ``path``. ``path`` must
    be a directory. If ``no_watch`` is True, any change of the files in the
    path do not trigger restart if ``--autoreload`` is enabled.
    """
    app = get_app()
    app.add_static_path(prefix, path)
    if not no_watch:
        watch_dir(path)


def get_app(*args, **kwargs):
    """Get root Application object."""
    return module.get_app()


def set_server_type(type):
    """Set server type to document and handle websocket connection.

    By default, if aiohttp is available wdom uses aiohttp for server. Otherwise
    use tornado.
    """
    global module
    if type == 'aiohttp':
        module = importlib.import_module('wdom.server._aiohttp')
    elif type == 'tornado':
        module = importlib.import_module('wdom.server._tornado')
    else:
        raise ValueError(
            '{0} is not supported now. Available server types are:'
            ' aiohttp, tornado'.format(type)
        )


def start_server(app: Optional[module.Application] = None,
                 browser: Optional[str] = None,
                 address: Optional[str] = None,
                 check_time: Optional[int] = 500,
                 **kwargs):
    """Start web server."""
    # Add application's static files directory
    from wdom.document import get_document
    add_static_path('_static', static_dir)
    doc = get_document()
    if os.path.exists(doc.tempdir):
        add_static_path('tmp', doc.tempdir, no_watch=True)
    if doc._autoreload or config.autoreload or config.debug:
        install_asyncio()
        autoreload.start(check_time=check_time)
    global _server
    _server = module.start_server(**kwargs)
    logger.info('Start server on {0}:{1:d}'.format(
        _server.address, _server.port))

    if config.open_browser:
        open_browser('http://{}:{}/'.format(_server.address, _server.port),
                     browser or config.browser)
    return _server


def stop_server(server=None):
    """Terminate web server."""
    module.stop_server(server or _server)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.testing import TestCase

from wdom.node import Node
from wdom.parser import DocumentParser, FragmentParser


class TestDocumentParser(TestCase):
    def setUp(self):
        self.parser = DocumentParser()

    def test_empty(self):
        self.parser.feed('')
        doc = self.parser.root
        self.assertEqual(doc.length, 2)
        self.assertEqual(doc.nodeType, Node.DOCUMENT_NODE)
        self.assertEqual(doc.doctype.nodeType, Node.DOCUMENT_TYPE_NODE)
        self.assertEqual(doc.html.length, 2)

    def test_single_tag(self):
        self.parser.feed('<h1>title</h1>')
        doc = self.parser.root
        self.assertEqual(doc.html.length, 2)
        self.assertEqual(doc.body.length, 2)
        # first child is <script>
        self.assertEqual(doc.body.lastChild.tagName, 'H1')

    def test_full_docc(self):
        self.parser.feed(
            '<!DOCTYPE html><html><head><title>Title</title></head>'
            '<body><h1>title</h1></body></html>')
        doc = self.parser.root
        self.assertEqual(doc.html.length, 2)
        self.assertEqual(doc.body.length, 2)
        # first child is <script>
        self.assertEqual(doc.body.lastChild.tagName, 'H1')


class TestFragmentParser(TestCase):
    def setUp(self):
        self.parser = FragmentParser()

    def test_empty(self):
        self.parser.feed('')
        self.assertEqual(self.parser.root.length, 0)
        self.assertEqual(self.parser.root.nodeType,
                         Node.DOCUMENT_FRAGMENT_NODE)

    def test_single_tag(self):
        self.parser.feed('<h1>test</h1>')
        self.assertEqual(self.parser.root.length, 1)
        child = self.parser.root.firstChild
        self.assertEqual(child.nodeType, Node.ELEMENT_NODE)
        self.assertEqual(child.textContent, 'test')
        self.assertEqual(child.textContent, 'test')

    def test_multi_tag_para(self):
        self.parser.feed('<h1>test1</h1><h2>test2</h2><h3>test3</h3>')
        self.assertEqual(self.parser.root.length, 3)
        child1 = self.parser.root.childNodes[0]
        child2 = self.parser.root.childNodes[1]
        child3 = self.parser.root.childNodes[2]
        self.assertEqual(child1.length, 1)
        self.assertEqual(child2.length, 1)
        self.assertEqual(child3.length, 1)
        self.assertEqual(child1.tagName, 'H1')
        self.assertEqual(child2.tagName, 'H2')
        self.assertEqual(child3.tagName, 'H3')
        self.assertEqual(child1.textContent, 'test1')
        self.assertEqual(child2.textContent, 'test2')
        self.assertEqual(child3.textContent, 'test3')

    def test_multi_tag_nest(self):
        self.parser.feed('<h1>test1<h2>test2<h3>test3</h3></h2></h1>')
        self.assertEqual(self.parser.root.length, 1)
        child1 = self.parser.root.childNodes[0]
        self.assertEqual(child1.length, 2)  # text and h2
        child2 = child1.lastChild
        self.assertEqual(child2.length, 2)  # text and h3
        child3 = child2.lastChild
        self.assertEqual(child3.length, 1)  # only text

        self.assertEqual(child1.textContent, 'test1test2test3')
        self.assertEqual(child2.textContent, 'test2test3')
        self.assertEqual(child3.textContent, 'test3')

    def test_multi_tag_para_nest(self):
        self.parser.feed(
            '<h1>test1<h2>test2</h2></h1><h3>test3<h4>test4</h4></h3>')
        self.assertEqual(self.parser.root.length, 2)

        child1 = self.parser.root.childNodes[0]
        child2 = child1.childNodes[1]
        child3 = self.parser.root.childNodes[1]
        child4 = child3.childNodes[1]

        self.assertEqual(child1.length, 2)  # text and h2
        self.assertEqual(child2.length, 1)  # only text
        self.assertEqual(child3.length, 2)  # text and h4
        self.assertEqual(child4.length, 1)  # only text

        self.assertEqual(child1.tagName, 'H1')
        self.assertEqual(child2.tagName, 'H2')
        self.assertEqual(child3.tagName, 'H3')
        self.assertEqual(child4.tagName, 'H4')

    def test_empty_tag(self):
        self.parser.feed('<h1><img src="a"></h1>')
        self.assertEqual(self.parser.root.length, 1)
        h1 = self.parser.root.firstChild
        self.assertEqual(h1.length, 1)
        img = h1.firstChild
        self.assertEqual(img.length, 0)
        self.assertEqual(img.tagName, 'IMG')

    def test_empty_tag_only(self):
        self.parser.feed('<img src="a">')
        self.assertEqual(self.parser.root.length, 1)
        img = self.parser.root.firstChild
        self.assertEqual(img.length, 0)
        self.assertEqual(img.tagName, 'IMG')

    def test_empty_tag_para(self):
        self.parser.feed('<h1><img src="a"><img src="b"></h1>')
        self.assertEqual(self.parser.root.length, 1)
        h1 = self.parser.root.firstChild
        self.assertEqual(h1.length, 2)
        img1 = h1.firstChild
        self.assertEqual(img1.length, 0)
        self.assertEqual(img1.tagName, 'IMG')
        self.assertEqual(img1.getAttribute('src'), 'a')
        img2 = h1.lastChild
        self.assertEqual(img2.length, 0)
        self.assertEqual(img2.tagName, 'IMG')
        self.assertEqual(img2.getAttribute('src'), 'b')

    def test_empty_tag_text(self):
        self.parser.feed('<h1>test1<img src="a">test2</h1>')
        self.assertEqual(self.parser.root.length, 1)
        h1 = self.parser.root.firstChild
        self.assertEqual(h1.length, 3)
        self.assertEqual(h1.textContent, 'test1test2')

        img = h1.childNodes[1]
        self.assertEqual(img.length, 0)

    def test_comment(self):
        self.parser.feed('<!--comment-->')
        self.assertEqual(self.parser.root.length, 1)
        c = self.parser.root.firstChild
        self.assertEqual(c.length, 7)
        self.assertEqual(c.html, '<!--comment-->')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import json
import logging

from wdom.interface import Event

logger = logging.getLogger(__name__)


def log_handler(level: str, message: str):
    """Handle logs from client (browser)."""
    message = 'JS: ' + str(message)
    if level == 'error':
        logger.error(message)
    elif level == 'warn':
        logger.warning(message)
    elif level == 'info':
        logger.info(message)
    elif level == 'debug':
        logger.debug(message)


def event_handler(msg: dict):
    """Handle events emitted on browser."""
    from wdom.document import getElementByRimoId
    e = Event(**msg.get('event'))
    _id = e.currentTarget.get('id')
    currentTarget = getElementByRimoId(_id)
    if currentTarget is None:
        logger.warning('No such element: rimo_id={}'.format(_id))
        return

    currentTarget.on_event_pre(e)
    e.currentTarget = currentTarget
    e.target = getElementByRimoId(e.target.get('id'))
    e.currentTarget.dispatchEvent(e)


def response_handler(msg: dict):
    """Handle response sent by browser."""
    from wdom.document import getElementByRimoId
    id = msg.get('id')
    elm = getElementByRimoId(id)
    if elm:
        elm.on_response(msg)
    else:
        logger.warning('No such element: rimo_id={}'.format(id))


def on_websocket_message(message):
    """Handle messages from browser."""
    msg = json.loads(message)
    _type = msg.get('type')
    if _type == 'log':
        log_handler(msg.get('level'), msg.get('message'))
    elif _type == 'event':
        event_handler(msg)
    elif _type == 'response':
        response_handler(msg)
    else:
        raise ValueError('unkown message type: {}'.format(message))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.tag import NewTagClass as NewTag
from wdom.tag import *

name = 'SkyBlue'
project_url = 'http://stanko.github.io/skyblue/'
project_repository = 'https://github.com/Stanko/skyblue'
license = 'MIT License'
license_url = 'https://github.com/Stanko/skyblue/blob/gh-pages/LICENSE.md'

css_files = [
    '_static/css/skyblue.min.css',
]


Button = NewTag('Button', 'a', bases=A, class_='btn')
DefaultButton = NewTag('DefaultButton', 'a', Button, class_='btn-dark', is_='default-button')
PrimaryButton = NewTag('PrimaryButton', 'a', Button, is_='primary-button')
SecondaryButton = NewTag('SecondaryButton', 'a', Button, class_='btn-light', is_='secondary-button')
SuccessButton = NewTag('SuccessButton', 'a', Button, class_='btn-success', is_='success-button')
InfoButton = NewTag('InfoButton', 'a', Button, class_='btn-empty', is_='info-button')
WarningButton = NewTag('WarningButton', 'a', Button, class_='btn-warning', is_='warning-button')
DangerButton = NewTag('DangerButton', 'a', Button, class_='btn-error', is_='danger-button')
ErrorButton = NewTag('ErrorButton', 'a', Button, class_='btn-error', is_='error-button')
LinkButton = NewTag('LinkButton', 'a', Button, class_='btn-empty btn-success', is_='link-button')

Input = NewTag('Input', 'input', Input, class_='form-control', type_='text')
TextInput = NewTag('TextInput', 'input', TextInput, class_='form-control')
Textarea = NewTag('Textarea', 'textarea', Textarea, class_='form-control')
Select = NewTag('Select', 'select', Select, class_='form-control')

Table = NewTag('Table', 'table', Table, class_='table')

Container = NewTag('Container', 'div', Container, class_='container')
Wrapper = NewTag('Wrapper', 'div', Wrapper, class_='container')
Row = NewTag('Row', 'div', Row, class_='row')

Col = NewTag('Col', 'div', Col, class_='col')
Col1 = NewTag('Col1', 'div', Col1, class_='col xs-1 sm-1 md-1 lg-1 xl-1')
Col2 = NewTag('Col2', 'div', Col2, class_='col xs-2 sm-2 md-2 lg-2 xl-2')
Col3 = NewTag('Col3', 'div', Col3, class_='col xs-3 sm-3 md-3 lg-3 xl-3')
Col4 = NewTag('Col4', 'div', Col4, class_='col xs-4 sm-4 md-4 lg-4 xl-4')
Col5 = NewTag('Col5', 'div', Col5, class_='col xs-5 sm-5 md-5 lg-5 xl-5')
Col6 = NewTag('Col6', 'div', Col6, class_='col xs-6 sm-6 md-6 lg-6 xl-6')
Col7 = NewTag('Col7', 'div', Col7, class_='col xs-7 sm-7 md-7 lg-7 xl-7')
Col8 = NewTag('Col8', 'div', Col8, class_='col xs-8 sm-8 md-8 lg-8 xl-8')
Col9 = NewTag('Col9', 'div', Col9, class_='col xs-9 sm-9 md-9 lg-9 xl-9')
Col10 = NewTag('Col10', 'div', Col10, class_='col xs-10 sm-10 md-10 lg-10 xl-10')
Col11 = NewTag('Col11', 'div', Col11, class_='col xs-11 sm-11 md-11 lg-11 xl-11')
Col12 = NewTag('Col12', 'div', Col12, class_='col xs-12 sm-12 md-12 lg-12 xl-12')

extended_classes = [
    Button,
    DefaultButton,
    PrimaryButton,
    SecondaryButton,
    SuccessButton,
    InfoButton,
    WarningButton,
    DangerButton,
    ErrorButton,
    LinkButton,
    Input,
    TextInput,
    Textarea,
    Select,
    Table,
    Container,
    Wrapper,
    Row,
    Col,
    Col1,
    Col2,
    Col3,
    Col4,
    Col5,
    Col6,
    Col7,
    Col8,
    Col9,
    Col10,
    Col11,
    Col12,
]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from os import path

sys.path.append(path.dirname(path.dirname(__file__)))

from xml.etree.ElementTree import HTML_EMPTY
from html.parser import HTMLParser

from wdom.document import Document
from wdom.node import DocumentFragment, Comment
from wdom.web_node import WebElement


class DocumentParser(HTMLParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.root = Document()
        self.elm = self.root.body

    def handle_decl(self, decl):
        if decl.startswith('DOCTYPE'):
            self.root.doctype.name = decl.split()[-1]

    def handle_starttag(self, tag, attrs):
        if tag == 'html':
            self.elm = self.root.html
        elif tag == 'head':
            self.elm = self.root.head
        elif tag == 'body':
            self.elm = self.root.body
        else:
            elm = WebElement(tag, parent=self.elm, **dict(attrs))
            if self.elm:
                self.elm.append(elm)
            if tag not in HTML_EMPTY:
                self.elm = elm

    def handle_endtag(self, tag):
        self.elm = self.elm.parentNode

    def handle_data(self, data):
        _d = data.strip()
        if _d and self.elm:
            self.elm.append(_d)

    def handle_comment(self, comment:str):
        self.elm.append(Comment(comment))


class FragmentParser(HTMLParser):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.elm = DocumentFragment()
        self.root = self.elm

    def handle_starttag(self, tag, attrs):
        elm = WebElement(tag, parent=self.elm, **dict(attrs))
        if self.elm:
            self.elm.append(elm)
        if tag not in HTML_EMPTY:
            self.elm = elm

    def handle_endtag(self, tag):
        self.elm = self.elm.parentNode

    def handle_data(self, data):
        if data and self.elm:
            self.elm.append(data)

    def handle_comment(self, comment:str):
        self.elm.append(Comment(comment))


def parse_document(doc:str):
    parser = DocumentParser()
    parser.feet(doc)
    return parser.root


def parse_html(html:str):
    parser = FragmentParser()
    parser.feed(html)
    return parser.root



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
from os import path
from tempfile import NamedTemporaryFile
from copy import copy
import logging
import unittest
from importlib import reload
import subprocess

from nose_parameterized import parameterized

from wdom.misc import root_dir
from wdom.options import parse_command_line, config, set_loglevel
from wdom import tag
from wdom.themes import default as default_theme
from wdom.testing import TestCase

_argv = copy(sys.argv)
_config = copy(vars(config))


def reset_options():
    sys.argv = copy(_argv)
    for k, v in _config.items():
        setattr(config, k, v)


class TestOptions(TestCase):
    def tearDown(self):
        reset_options()
        super().tearDown()

    def test_default_loglevel(self):
        parse_command_line()
        set_loglevel()
        logger = logging.getLogger('wdom')
        self.assertEqual(logger.getEffectiveLevel(), logging.INFO)

    @parameterized.expand([
        ('debug', logging.DEBUG),
        ('info', logging.INFO),
        ('warn', logging.WARN),
        ('error', logging.ERROR),
    ])
    def test_loglevel(self, level_name, level):
        sys.argv.extend(['--logging', level_name])
        parse_command_line()
        logger = logging.getLogger('wdom')
        self.assertEqual(logger.getEffectiveLevel(), level)
        reset_options()

    def test_debug_without_logging(self):
        sys.argv.extend(['--debug'])
        parse_command_line()
        logger = logging.getLogger('wdom')
        self.assertEqual(logger.getEffectiveLevel(), logging.DEBUG)

    def test_debug_with_logging(self):
        sys.argv.extend(['--debug', '--logging', 'warn'])
        parse_command_line()
        logger = logging.getLogger('wdom')
        assert logger.getEffectiveLevel() == logging.WARN

    def test_unknown_args(self):
        src = '\n'.join([
            'import sys',
            'sys.path.append(\'{}\')'.format(path.dirname(root_dir)),
            'from wdom import options',
        ])
        with NamedTemporaryFile('w', dir=path.abspath(path.curdir)) as f:
            f.write(src)
            f.flush()  # save file
            cmd = [sys.executable, f.name, '--test-arg']
            proc = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, stderr=subprocess.STDOUT,
                universal_newlines=True,
            )
            proc.wait()
        result = proc.stdout.read()
        self.assertIn('usage: WDOM', result)
        self.assertIn('Unknown Argument', result)
        self.assertIn('--test-arg', result)


class TestThemeOption(unittest.TestCase):
    def tearDown(self):
        reset_options()
        super().tearDown()

    def test_no_theme(self):
        parse_command_line()
        with self.assertLogs('wdom.themes', 'INFO') as log:
            reload(default_theme)
        self.assertEqual(len(log.output), 1)
        self.assertTrue(log.records[0].msg.startswith('No theme'))
        self.assertTrue(log.records[0].msg.endswith('Use default theme.'))
        self.assertEqual(default_theme.Button, tag.Button)

    def test_bs(self):
        sys.argv.extend(['--theme', 'bootstrap3'])
        parse_command_line()
        with self.assertLogs('wdom.themes', 'INFO') as log:
            reload(default_theme)
        self.assertEqual(len(log.output), 1)
        self.assertEqual(log.records[0].msg, 'Use theme: bootstrap3')
        from wdom.themes import bootstrap3
        self.assertEqual(default_theme.Button, bootstrap3.Button)

    def test_unknown_theme(self):
        sys.argv.extend(['--theme', 'unknown'])
        parse_command_line()
        with self.assertLogs('wdom.themes', 'WARN') as log:
            reload(default_theme)
        self.assertEqual(len(log.output), 2)
        self.assertTrue(log.records[0].msg.startswith('Unknown theme'))
        self.assertTrue(log.records[1].msg.startswith('Available themes:'))
        self.assertIn('unknown', log.records[0].msg)
        self.assertIn('skeleton', log.records[1].msg)
        self.assertEqual(default_theme.Button, tag.Button)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from wdom.element import HTMLElement
from wdom.tag import Tag, DOMTokenList, NewTagClass, NestedTag
from wdom.window import customElements
from wdom.testing import TestCase


class TestTag(TestCase):
    '''Test for Basic Dom implementation'''
    def setUp(self):
        customElements.clear()
        self.tag = Tag()
        self.c1 = Tag(c="1")
        self.c2 = Tag(c="2")

    def test_name(self):
        self.assertEqual(self.tag.tag, 'tag')
        self.assertEqual(self.tag.tagName, 'TAG')
        self.assertEqual(self.tag.localName, 'tag')

        class A(Tag):
            tag = 'Atag'
        a = A()
        self.assertEqual(a.tag, 'Atag')
        self.assertEqual(a.tagName, 'ATAG')
        self.assertEqual(a.localName, 'atag')

    def test_tag_string(self):
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')

    def test_attr_init(self):
        tag = Tag(attrs={'src': 'a'})
        self.assertRegex(tag.html, '<tag src="a" rimo_id="\d+"></tag>')
        tag.removeAttribute('src')
        self.assertRegex(tag.html, '<tag rimo_id="\d+"></tag>')

    def test_attr_atomic(self):
        # test add tag-attr
        self.tag['a'] = 'b'
        self.assertEqual(self.tag['a'], 'b')
        self.assertIn('a="b"', self.tag.html)
        self.assertRegex(self.tag.start_tag, '<tag rimo_id="\d+" a="b">')
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+" a="b"></tag>')
        del self.tag['a']
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')

    def test_attr_addremove(self):
        self.assertTrue(self.tag.hasAttributes())  # has id
        self.assertFalse(self.tag.hasAttribute('a'))
        self.tag.setAttribute('a', 'b')
        self.assertTrue(self.tag.hasAttributes())
        self.assertTrue(self.tag.hasAttribute('a'))
        self.assertIsFalse(self.tag.hasAttribute('b'))
        self.assertEqual('b', self.tag.getAttribute('a'))
        self.assertRegex(self.tag.html, r'<tag rimo_id="\d+" a="b"></tag>')
        self.assertEqual(self.tag.getAttribute('a'), 'b')
        self.tag.removeAttribute('a')
        self.assertTrue(self.tag.hasAttributes())
        self.assertFalse(self.tag.hasAttribute('a'))
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')
        self.assertIsNone(self.tag.getAttribute('aaaa'))

    def test_attr_multi(self):
        self.tag.setAttribute('c', 'd')
        self.tag.setAttribute('e', 'f')
        self.assertIn('c="d" e="f"', self.tag.html)

    def test_attr_overwrite(self):
        self.tag.setAttribute('c', 'd')
        self.tag.setAttribute('e', 'f')
        self.tag.setAttribute('c', 'new_d')
        self.assertNotIn('c="d"', self.tag.html)
        self.assertIn('c="new_d"', self.tag.html)
        self.assertIn('e="f"', self.tag.html)

    def test_child_addremove(self):
        self.assertIsFalse(self.tag.hasChildNodes())
        self.tag.appendChild(self.c1)
        self.assertIsTrue(self.tag.hasChildNodes())
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+"><tag c="1" rimo_id="\d+"></tag></tag>',
        )
        self.assertIn(self.c1, self.tag)
        self.tag.removeChild(self.c1)
        self.assertIsFalse(self.tag.hasChildNodes())
        self.assertNotIn(self.c1, self.tag)
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')

    def test_child_exception(self) -> None:
        with self.assertRaises(ValueError):
            self.tag.removeChild(Tag())
        with self.assertRaises(ValueError):
            self.tag.replaceChild(Tag(), Tag())

    def test_first_last_child(self):
        self.assertIsNone(self.tag.firstChild)
        self.assertIsNone(self.tag.lastChild)
        self.tag.appendChild(self.c1)
        self.assertIs(self.tag.firstChild, self.c1)
        self.assertIs(self.tag.lastChild, self.c1)
        self.tag.appendChild(self.c2)
        self.assertIs(self.tag.firstChild, self.c1)
        self.assertIs(self.tag.lastChild, self.c2)

    def test_child_deep(self):
        self.tag.appendChild(self.c1)
        self.c1.appendChild(self.c2)
        self.assertNotIn(self.c2, self.tag)
        self.assertIn(self.c2, self.c1)
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+"><tag c="1" rimo_id="\d+">'
            '<tag c="2" rimo_id="\d+"></tag></tag></tag>',
        )

    def test_child_nodes(self):
        self.tag.appendChild(self.c1)
        self.tag.appendChild(self.c2)
        self.assertEqual(len(self.tag.childNodes), 2)
        self.assertIs(self.tag.childNodes[0], self.c1)
        self.assertIs(self.tag.childNodes[1], self.c2)

    def test_child_replace(self):
        self.tag.append(self.c1)
        self.assertIn(self.c1, self.tag)
        self.assertNotIn(self.c2, self.tag)
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+"><tag c="1" rimo_id="\d+"></tag></tag>',
        )

        self.tag.replaceChild(self.c2, self.c1)
        self.assertNotIn(self.c1, self.tag)
        self.assertIn(self.c2, self.tag)
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+"><tag c="2" rimo_id="\d+"></tag></tag>',
        )

    def test_text_addremove(self):
        self.tag.textContent = 'text'
        self.assertIsTrue(self.tag.hasChildNodes())
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+">text</tag>')
        # self.assertIn('text', self.tag)
        self.assertEqual(self.tag[0].parentNode, self.tag)

        self.tag.textContent = ''
        self.assertIsFalse(self.tag.hasChildNodes())
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')

    def test_textcontent(self):
        self.assertEqual(self.tag.textContent, '')
        self.tag.textContent = 'a'
        self.assertEqual(self.tag.textContent, 'a')
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+">a</tag>')
        self.tag.textContent = 'b'
        self.assertEqual(self.tag.textContent, 'b')
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+">b</tag>')

    def test_textcontent_child(self):
        self.tag.textContent = 'a'
        self.tag.appendChild(self.c1)
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+">a<tag c="1" rimo_id="\d+"></tag></tag>',
        )
        self.c1.textContent = 'c1'
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+">a<tag c="1" rimo_id="\d+">c1</tag></tag>',
        )
        self.assertEqual('ac1', self.tag.textContent)
        self.tag.textContent = 'b'
        self.assertEqual(self.tag.length, 1)
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+">b</tag>')
        self.assertIsNone(self.c1.parentNode)

    def test_closing_tag(self):
        class Img(Tag):
            tag = 'img'
        img = Img()
        self.assertRegex(img.html, '<img rimo_id="\d+">')
        img.setAttribute('src', 'a')
        self.assertRegex(img.html, '<img rimo_id="\d+" src="a">')

    def _test_shallow_copy(self, clone):
        self.assertIsTrue(self.tag.hasChildNodes())
        self.assertIsFalse(clone.hasChildNodes())
        self.assertEqual(len(clone), 0)
        self.assertRegex(
            clone.html,
            '<tag rimo_id="\d+" src="a" class="b"></tag>',
        )

        self.assertIsTrue(clone.hasAttributes())
        self.assertEqual(clone.getAttribute('src'), 'a')
        clone.setAttribute('src', 'b')
        self.assertEqual(clone.getAttribute('src'), 'b')
        self.assertEqual(self.tag.getAttribute('src'), 'a')

        self.assertIsTrue(clone.hasClass('b'))
        self.assertEqual(clone.getAttribute('class'), 'b')
        clone.setAttribute('class', 'c')
        self.assertEqual(clone.getAttribute('class'), 'c')
        self.assertEqual(self.tag.getAttribute('class'), 'b')

        clone.append(self.c2)
        self.assertIsTrue(clone.hasChildNodes())
        self.assertIn(self.c2, clone)
        self.assertNotIn(self.c2, self.tag)

    def test_copy(self):
        from copy import copy
        self.tag.appendChild(self.c1)
        self.tag.setAttribute('src', 'a')
        self.tag.setAttribute('class', 'b')
        clone = copy(self.tag)
        self._test_shallow_copy(clone)

    def test_clone_node_sharrow(self):
        self.tag.appendChild(self.c1)
        self.tag.setAttribute('src', 'a')
        self.tag.setAttribute('class', 'b')
        clone = self.tag.cloneNode()
        self._test_shallow_copy(clone)

        clone2 = self.tag.cloneNode(deep=False)
        self._test_shallow_copy(clone2)

    def _test_deep_copy(self, clone):
        self.assertIsTrue(clone.hasChildNodes())
        self.assertEqual(len(clone), 1)
        self.assertIn(self.c1, self.tag)
        self.assertNotIn(self.c1, clone)

        self.c1.setAttribute('src', 'b')
        self.assertEqual(self.c1.getAttribute('src'), 'b')
        self.assertIsNone(clone[0].getAttribute('src'))

        self.c1.setAttribute('class', 'c')
        self.assertEqual(self.c1.getAttribute('class'), 'c')
        self.assertIsNone(clone[0].getAttribute('class'))

        clone.append(self.c2)
        self.assertEqual(len(clone), 2)
        self.assertEqual(len(self.tag), 1)

    def test_deepcopy(self):
        from copy import deepcopy
        self.tag.append(self.c1)
        self.tag.setAttribute('src', 'a')
        self.tag.setAttribute('class', 'b')
        clone = deepcopy(self.tag)
        self._test_deep_copy(clone)

    def test_clone_node_deep(self):
        self.tag.append(self.c1)
        self.tag.setAttribute('src', 'a')
        self.tag.setAttribute('class', 'b')
        clone = self.tag.cloneNode(deep=True)
        self._test_deep_copy(clone)

    def test_siblings(self):
        self.tag.appendChild(self.c1)
        self.tag.appendChild(self.c2)
        self.assertIsNone(self.tag.previousSibling)
        self.assertIsNone(self.tag.nextSibling)
        self.assertIsNone(self.c1.previousSibling)
        self.assertIs(self.c2.previousSibling, self.c1)
        self.assertIs(self.c1.nextSibling, self.c2)
        self.assertIsNone(self.c2.nextSibling)

    def test_get_elements_by_tagname(self):
        A = NewTagClass('A', 'a')
        B = NewTagClass('B', 'b')
        a1 = A(src='a1')
        a2 = A(src='a2')
        b1 = B(src='b1')
        b2 = B(src='b2')
        self.tag.appendChild(a1)
        self.tag.appendChild(a2)
        self.tag.appendChild(b1)
        b1.appendChild(b2)

        a_list = self.tag.getElementsByTagName('a')
        self.assertEqual(len(a_list), 2)
        self.assertIs(a_list[0], a1)
        self.assertIs(a_list[1], a2)

        b_list = self.tag.getElementsByTagName('b')
        self.assertEqual(len(b_list), 2)
        self.assertIs(b_list[0], b1)
        self.assertIs(b_list[1], b2)

        b_sub_list = b1.getElementsByTagName('b')
        self.assertEqual(len(b_sub_list), 1)
        self.assertIs(b_sub_list[0], b2)

    class NewTag(Tag):
        tag = 'new-tag'

    def test_custom_tag(self):
        self.tag.innerHTML = '<new-tag></new-tag>'
        self.assertEqual(type(self.tag.firstChild), HTMLElement)
        self.assertFalse(self.tag.firstChild._registered)
        customElements.define('new-tag', self.NewTag)
        self.assertEqual(type(self.tag.firstChild), self.NewTag)
        self.assertTrue(self.tag.firstChild._registered)

    def test_custom_tag_registered(self):
        customElements.define('new-tag', self.NewTag)
        self.c1.innerHTML = '<new-tag></new-tag>'
        self.assertTrue(isinstance(self.c1.firstChild, self.NewTag))
        self.assertTrue(self.c1.firstChild._registered)

    class ExtendTag(Tag):
        tag = 'a'
        is_ = 'new-a'

    def test_custom_tag_is(self):
        self.tag.innerHTML = '<a is="new-a"></a>'
        self.assertEqual(type(self.tag.firstChild), HTMLElement)
        self.assertFalse(self.tag.firstChild._registered)
        customElements.define('new-a', self.NewTag, {'extends': 'a'})
        self.assertEqual(type(self.tag.firstChild), self.NewTag)
        self.assertTrue(self.tag.firstChild._registered)

    def test_custom_tag_is_registered(self):
        customElements.define('new-a', self.ExtendTag, {'extends': 'a'})
        self.tag.innerHTML = '<a is="new-a"></a>'
        self.assertEqual(type(self.tag.firstChild), self.ExtendTag)
        self.assertTrue(self.tag.firstChild._registered)

        # test unregistered `is`
        self.tag.innerHTML = '<a is="new-b"></a>'
        self.assertEqual(type(self.tag.firstChild), HTMLElement)
        self.assertFalse(self.tag.firstChild._registered)

    def test_custom_tag_define_by_class(self):
        customElements.define(self.NewTag)
        self.tag.innerHTML = '<new-tag></new-tag>'
        self.assertEqual(type(self.tag.firstChild), self.NewTag)

    def test_custom_tag_define_by_class_is(self):
        customElements.define(self.ExtendTag)
        self.tag.innerHTML = '<a is="new-a"></a>'
        self.assertEqual(type(self.tag.firstChild), self.ExtendTag)


class TestNestedTag(TestCase):
    def setUp(self):
        class Inner(Tag):
            tag = 'in'

        class Outer(NestedTag):
            tag = 'out'
            inner_tag_class = Inner
        self.inner_cls = Inner
        self.outer_cls = Outer
        self.tag = self.outer_cls()
        self.c1 = Tag(src='c1')
        self.c2 = Tag(src='c2')
        self.c3 = Tag(src='c3')
        self.c4 = Tag(src='c4')

    def test_tag(self):
        tag = self.outer_cls()
        self.assertRegex(tag.html,
                         r'<out rimo_id="\d+"><in rimo_id="\d+"></in></out>')

    def test_creation(self):
        c = Tag(src='c1')
        tag = self.outer_cls(c, src='out')
        self.assertEqual(
            tag.html_noid,
            '<out src="out"><in><tag src="c1"></tag></in></out>',
        )

    def test_child_node(self):
        self.assertIsNone(self.tag.firstChild)
        self.assertIsNone(self.tag.lastChild)
        self.assertIsNone(self.tag.firstElementChild)
        self.assertIsNone(self.tag.lastElementChild)
        self.tag.appendChild(self.c1)
        self.assertIs(self.tag.firstChild, self.c1)
        self.assertIs(self.tag.lastChild, self.c1)
        self.assertIs(self.tag.firstElementChild, self.c1)
        self.assertIs(self.tag.lastElementChild, self.c1)
        self.assertIsNone(self.c1.previousSibling)
        self.assertIsNone(self.c1.previousElementSibling)
        self.assertIsNone(self.c1.nextSibling)
        self.assertIsNone(self.c1.nextElementSibling)

    def test_addremove_child(self):
        self.tag.append(self.c1, self.c2)
        self.assertEqual(
            self.tag.html_noid,
            '<out><in><tag src="c1"></tag><tag src="c2"></tag></in></out>',
        )
        self.tag.removeChild(self.c1)
        self.assertEqual(
            self.tag.html_noid, '<out><in><tag src="c2"></tag></in></out>')
        self.tag.replaceChild(self.c1, self.c2)
        self.assertEqual(
            self.tag.html_noid, '<out><in><tag src="c1"></tag></in></out>')
        self.c1.replaceWith(self.c2)
        self.assertEqual(
            self.tag.html_noid, '<out><in><tag src="c2"></tag></in></out>')
        self.c2.remove()
        self.assertEqual(
            self.tag.html_noid, '<out><in></in></out>')

    def test_insert(self):
        self.tag.append(self.c1, self.c2)
        self.tag.insertBefore(self.c3, self.c2)
        self.assertEqual(
            self.tag.html_noid,
            '<out><in><tag src="c1"></tag><tag src="c3"></tag>'
            '<tag src="c2"></tag></in></out>',
        )
        self.tag.prepend(self.c4)
        self.assertEqual(
            self.tag.html_noid,
            '<out><in><tag src="c4"></tag><tag src="c1"></tag>'
            '<tag src="c3"></tag><tag src="c2"></tag></in></out>',
        )
        self.tag.empty()
        self.assertEqual(self.tag.html_noid, '<out><in></in></out>')
        self.assertFalse(self.tag.hasChildNodes())
        self.tag.append(self.c1)
        self.c1.before(self.c2)
        self.assertEqual(
            self.tag.html_noid,
            '<out><in><tag src="c2"></tag><tag src="c1"></tag></in></out>',
        )
        self.c1.after(self.c3)
        self.assertEqual(
            self.tag.html_noid,
            '<out><in><tag src="c2"></tag><tag src="c1"></tag>'
            '<tag src="c3"></tag></in></out>',
        )

    def test_inner_content(self):
        self.assertEqual(self.tag.textContent, '')
        self.assertEqual(self.tag.innerHTML, '')
        self.tag.textContent = 'test'
        self.assertEqual(self.tag.textContent, 'test')
        self.assertEqual(self.tag.html_noid, '<out><in>test</in></out>')
        self.tag.append(self.c1)
        self.assertEqual(
            self.tag.html_noid,
            '<out><in>test<tag src="c1"></tag></in></out>',
        )
        self.tag.textContent = 'test2'
        self.assertEqual(self.tag.textContent, 'test2')
        self.assertEqual(self.tag.html_noid, '<out><in>test2</in></out>')
        self.tag.innerHTML = '<a></a>'
        self.assertEqual(self.tag.html_noid, '<out><in><a></a></in></out>')


class TestClassList(TestCase):
    def setUp(self):
        self.cl = DOMTokenList(self)

    def test_addremove(self):
        self.assertIsFalse(bool(self.cl))
        self.assertEqual(len(self.cl), 0)
        self.cl.add('a')
        self.assertIsTrue(bool(self.cl))
        self.assertEqual(len(self.cl), 1)
        self.assertIn('a', self.cl)
        self.assertEqual('a', self.cl.toString())
        self.cl.add('b')
        self.assertIsTrue(bool(self.cl))
        self.assertEqual(len(self.cl), 2)
        self.assertIn('a', self.cl)
        self.assertIn('b', self.cl)
        self.assertEqual('a b', self.cl.toString())
        self.cl.remove('a')
        self.assertIsTrue(bool(self.cl))
        self.assertEqual(len(self.cl), 1)
        self.assertNotIn('a', self.cl)
        self.assertIn('b', self.cl)
        self.assertEqual('b', self.cl.toString())


class TestTagBase(TestCase):
    def setUp(self):
        self.tag = Tag()
        self.c1 = Tag()
        self.c2 = Tag()

    def test_class_addremove(self):
        self.assertIsFalse(self.tag.hasClasses())
        self.assertIsFalse(self.tag.hasClass('a'))
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')
        self.tag.addClass('a')
        self.assertIsTrue(self.tag.hasClasses())
        self.assertIsTrue(self.tag.hasClass('a'))
        self.assertIsFalse(self.tag.hasClass('b'))
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+" class="a"></tag>')
        self.tag.removeClass('a')
        self.assertIsFalse(self.tag.hasClasses())
        self.assertIsFalse(self.tag.hasClass('a'))
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')

    def test_class_in_init(self) -> None:
        tag = Tag(class_='a')
        self.assertIsTrue(tag.hasClass('a'))
        self.assertIsTrue(tag.hasClasses())
        self.assertRegex(tag.html, '<tag rimo_id="\d+" class="a"></tag>')
        tag.removeClass('a')
        self.assertIsFalse(tag.hasClass('a'))
        self.assertIsFalse(tag.hasClasses())
        self.assertRegex(tag.html, '<tag rimo_id="\d+"></tag>')

    def test_class_addremove_multi(self):
        self.tag.addClass('a', 'b', 'c')
        self.assertIsTrue(self.tag.hasClasses())
        self.assertIsTrue(self.tag.hasClass('a'))
        self.assertIsTrue(self.tag.hasClass('b'))
        self.assertIsTrue(self.tag.hasClass('c'))
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+" class="a b c"></tag>',
        )
        self.tag.removeClass('a', 'c')
        self.assertIsTrue(self.tag.hasClasses())
        self.assertIsFalse(self.tag.hasClass('a'))
        self.assertIsTrue(self.tag.hasClass('b'))
        self.assertIsFalse(self.tag.hasClass('c'))
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+" class="b"></tag>')

    def test_class_addremove_multi_string(self):
        with self.assertRaises(ValueError):
            self.tag.addClass('a b')

    def test_class_getset(self) -> None:
        self.assertEqual(self.tag['class'], None)
        self.tag.addClass('a')
        self.assertEqual(self.tag['class'], 'a')
        self.tag['class'] = 'b'
        self.assertEqual(self.tag['class'], 'b')
        self.assertIsFalse(self.tag.hasClass('a'))
        self.assertIsTrue(self.tag.hasClass('b'))

    def test_class_remove_error(self) -> None:
        with self.assertLogs('wdom.tag', 'WARNING'):
            self.tag.removeClass('a')

    def test_type_class(self) -> None:
        class A(Tag):
            tag = 'input'
            type_ = 'button'
        a = A()
        self.assertRegex(a.html, '<input type="button" rimo_id="\d+">')

    def test_type_init(self) -> None:
        a = Tag(type='button')
        self.assertRegex(a.html, '<tag type="button" rimo_id="\d+"></tag>')

    def test_type_attr(self) -> None:
        a = Tag()
        a.setAttribute('type', 'checkbox')
        self.assertRegex(a.html, '<tag rimo_id="\d+" type="checkbox"></tag>')

    def test_type_setter(self) -> None:
        class Check(Tag):
            type_ = 'checkbox'
        a = Check()
        b = Check()
        c = Check()
        b['type'] = 'radio'
        c.setAttribute('type', 'text')
        d = Check()
        self.assertRegex(a.html, '<tag type="checkbox" rimo_id="\d+"></tag>')
        self.assertRegex(b.html, '<tag type="radio" rimo_id="\d+"></tag>')
        self.assertRegex(c.html, '<tag type="text" rimo_id="\d+"></tag>')
        self.assertRegex(d.html, '<tag type="checkbox" rimo_id="\d+"></tag>')

    def test_hidden(self):
        self.tag.show()
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')
        self.tag.hide()
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+" hidden></tag>')
        self.tag.show()
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+"></tag>')

    def test_clone_node_sharrow_class(self):
        self.tag.appendChild(self.c1)
        self.tag.addClass('a')
        clone = self.tag.cloneNode()
        self.assertRegex(clone.html, '<tag rimo_id="\d+" class="a"></tag>')

        clone.removeClass('a')
        self.assertRegex(clone.html, '<tag rimo_id="\d+"></tag>')
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+"></tag></tag>',
        )

        clone.addClass('b')
        self.assertRegex(clone.html, '<tag rimo_id="\d+" class="b"></tag>')
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+"></tag></tag>',
        )

    def test_clone_node_sharrow_hidden(self):
        self.tag.hide()
        clone = self.tag.cloneNode()
        self.assertRegex(clone.html, '<tag rimo_id="\d+" hidden></tag>')
        clone.show()
        self.assertRegex(self.tag.html, '<tag rimo_id="\d+" hidden></tag>')
        self.assertRegex(clone.html, '<tag rimo_id="\d+"></tag>')

    def test_clone_node_deep_class(self):
        self.tag.appendChild(self.c1)
        self.tag.addClass('a')
        self.c1.addClass('b')
        clone = self.tag.cloneNode(deep=True)
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+" class="b">'
            '</tag></tag>',
        )
        self.assertRegex(
            clone.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+" class="b">'
            '</tag></tag>',
        )

        clone.childNodes[0].removeClass('b')
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+" class="b">'
            '</tag></tag>',
        )
        self.assertRegex(
            clone.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+"></tag></tag>',
        )

        self.c1.removeClass('b')
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+"></tag></tag>',
        )
        self.assertRegex(
            clone.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+"></tag></tag>',
        )

        clone.addClass('c')
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+"></tag></tag>',
        )
        self.assertRegex(
            clone.html,
            '<tag rimo_id="\d+" class="a c"><tag rimo_id="\d+"></tag></tag>',
        )

        clone.removeClass('a')
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+" class="a"><tag rimo_id="\d+"></tag></tag>',
        )
        self.assertRegex(
            clone.html,
            '<tag rimo_id="\d+" class="c"><tag rimo_id="\d+"></tag></tag>',
        )

    def test_clone_node_deep_hidden(self):
        self.tag.appendChild(self.c1)
        self.c1.hide()
        clone = self.tag.cloneNode(deep=True)
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+"><tag rimo_id="\d+" hidden></tag></tag>',
        )
        self.assertRegex(
            clone.html,
            '<tag rimo_id="\d+"><tag rimo_id="\d+" hidden></tag></tag>',
        )

        self.c1.show()
        self.assertRegex(
            self.tag.html,
            '<tag rimo_id="\d+"><tag rimo_id="\d+"></tag></tag>',
        )
        self.assertRegex(
            clone.html,
            '<tag rimo_id="\d+"><tag rimo_id="\d+" hidden></tag></tag>',
        )

    def test_class_of_class(self):
        class A(Tag):
            tag = 'a'
            class_ = 'a1'
        self.assertEqual(A.get_class_list().toString(), 'a1')
        a = A()
        self.assertRegex(a.html, '<a rimo_id="\d+" class="a1"></a>')
        a.addClass('a2')
        self.assertRegex(a.html, '<a rimo_id="\d+" class="a1 a2"></a>')
        with self.assertLogs('wdom.tag', 'WARNING'):
            a.removeClass('a1')
        self.assertRegex(a.html, '<a rimo_id="\d+" class="a1 a2"></a>')

    def test_classes_multiclass(self):
        class A(Tag):
            tag = 'a'
            class_ = 'a1 a2'
        self.assertEqual(A.get_class_list().toString(), 'a1 a2')
        a = A()
        a.addClass('a3', 'a4')
        self.assertRegex(a.html, '<a rimo_id="\d+" class="a1 a2 a3 a4"></a>')

    def test_classes_inherit_class(self):
        class A(Tag):
            tag = 'a'
            class_ = 'a1 a2'

        class B(A):
            tag = 'b'
            class_ = 'b1 b2'

        self.assertEqual(B.get_class_list().toString(), 'a1 a2 b1 b2')
        b = B()
        b.addClass('b3')
        self.assertRegex(
            b.html,
            '<b rimo_id="\d+" class="a1 a2 b1 b2 b3"></b>',
        )

    def test_classes_notinherit_class(self):
        class A(Tag):
            tag = 'a'
            class_ = 'a1 a2'

        class B(A):
            tag = 'b'
            class_ = 'b1 b2'
            inherit_class = False

        self.assertEqual(B.get_class_list().toString(), 'b1 b2')
        b = B()
        b.addClass('b3')
        self.assertRegex(b.html, '<b rimo_id="\d+" class="b1 b2 b3"></b>')

        class C(B):
            tag = 'c'
            class_ = 'c1 c2'
        self.assertEqual(C.get_class_list().toString(), 'b1 b2 c1 c2')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
import re
from typing import Union
from asyncio import coroutine
from weakref import WeakValueDictionary

from wdom.interface import Event
from wdom.node import Node
from wdom.webif import WebIF
from wdom.element import HTMLElement

logger = logging.getLogger(__name__)
remove_id_re = re.compile(r' rimo_id="\d+"')


class WebElement(HTMLElement, WebIF):
    _elements_with_rimo_id = WeakValueDictionary()

    @property
    def rimo_id(self) -> str:
        return self.getAttribute('rimo_id') or ''

    @rimo_id.setter
    def rimo_id(self, id: Union[str, int]):
        self.setAttribute('rimo_id', id)

    def __init__(self, *args, parent=None, rimo_id=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.rimo_id = rimo_id or str(id(self))
        self.addEventListener('mount', self._on_mount)
        if parent:
            parent.appendChild(self)

    def __copy__(self) -> HTMLElement:
        clone = super().__copy__()
        if clone.rimo_id == str(id(self)):
            # change automatically added id
            # overhead in __init__...
            clone.rimo_id = str(id(clone))
        return clone

    def _on_mount(self, *args, **kwargs):
        for event in self._listeners:
            self._add_event_listener_web(event=event)

    def _set_attribute(self, attr: str, value: str):
        if attr == 'rimo_id':
            if 'rimo_id' in self.attributes:
                # remove old reference to self
                self._elements_with_rimo_id.pop(self.rimo_id, None)
            # register this elements with new id
            self._elements_with_rimo_id[value] = self
        super()._set_attribute(attr, value)

    def _remove_web(self):
        self.js_exec('remove')

    def remove(self, *args, **kwargs):
        '''Remove this node from parent's DOM tree.
        '''
        self._remove_web()
        self._remove()

    def _empty_web(self):
        self.js_exec('empty')

    def empty(self):
        self._empty_web()
        self._empty()

    def _append_child_web(self, child: 'WebElement'):
        html = child.html if isinstance(child, Node) else str(child)
        self.js_exec('insertAdjacentHTML', 'beforeend', html)

    def appendChild(self, child: 'WebElement') -> Node:
        '''Append child node at the last of child nodes. If this instance is
        connected to the node on browser, the child node is also added to it.
        '''
        self._append_child_web(child)
        return self._append_child(child)

    def _insert_before_web(self, child: Node, ref_node: Node):
        html = child.html if isinstance(child, Node) else str(child)
        if isinstance(ref_node, WebElement):
            ref_node.js_exec('insertAdjacentHTML', 'beforebegin', html)
        else:
            index = self.index(ref_node)
            self.js_exec('insert', index, html)

    def insertBefore(self, child: Node, ref_node: Node) -> Node:
        '''Insert new child node before the reference child node. If the
        reference node is not a child of this node, raise ValueError. If this
        instance is connected to the node on browser, the child node is also
        added to it.
        '''
        self._insert_before_web(child, ref_node)
        return self._insert_before(child, ref_node)

    def _remove_child_web(self, child: Node):
        if child in self.childNodes:
            if isinstance(child, WebElement):
                self.js_exec('removeChildById', child.rimo_id)
            else:
                self.js_exec('removeChildByIndex', self.index(child))

    def removeChild(self, child: Node) -> Node:
        '''Remove the child node from this node. If the node is not a child
        of this node, raise ValueError.'''
        self._remove_child_web(child)
        return self._remove_child(child)

    def _replace_child_web(self, new_child: Node, old_child: Node):
        if isinstance(old_child, WebElement):
            self.js_exec('replaceChildById', new_child.html, old_child.rimo_id)
        else:
            # old_child will be Text Node
            index = old_child.parentNode.index(old_child)
            # Remove old_child before insert new child
            self._remove_child_web(old_child)
            self.js_exec('insert', index, new_child.html)

    def replaceChild(self, new_child, old_child) -> Node:
        self._replace_child_web(new_child, old_child)
        return self._replace_child(new_child, old_child)

    @coroutine
    def getBoundingClientRect(self):
        fut = yield from self.js_query('getBoundingClientRect')
        return fut

    @property
    def textContent(self) -> str:
        '''Return text contents of this node and all chid nodes. Any value is
        set to this property, all child nodes are removed and new value is set
        as a text node.
        '''
        return self._get_text_content()

    def _set_text_content_web(self, text: str):
        self.js_exec('textContent', self.textContent)

    @textContent.setter
    def textContent(self, text: str):
        self._set_text_content(text)
        self._set_text_content_web(text)

    def _set_inner_html_web(self, html: str):
        self.js_exec('innerHTML', html)

    @property
    def innerHTML(self) -> str:
        return self._get_inner_html()

    @innerHTML.setter
    def innerHTML(self, html: str):
        self._set_inner_html_web(html)
        self._set_inner_html(html)

    @property
    def html_noid(self) -> str:
        return remove_id_re.sub('', self.html)

    def click(self):
        if self.connected:
            self.js_exec('click')
        else:
            self._dispatch_event(Event('click'))

    def exec(self, script):
        self.js_exec('eval', script)

    # Window controll
    def scroll(self, x: int, y: int):
        self.js_exec('scroll', x, y)

    def scrollTo(self, x: int, y: int):
        self.js_exec('scrollTo', x, y)

    def scrollBy(self, x: int, y: int):
        self.js_exec('scrollBy', x, y)

    def scrollX(self):
        return self.js_query('scrollX')

    def scrollY(self):
        return self.js_query('scrollY')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import asyncio
from unittest.mock import MagicMock, call

from syncer import sync

from wdom.testing import TestCase
from wdom.event import Event, EventListener, EventTarget


class TestEventListener(TestCase):
    def setUp(self):
        self._cofunc_call_count = 0
        self._cofunc_calls = []

        @asyncio.coroutine
        def a(event):
            nonlocal self
            self._cofunc_call_count += 1
            self._cofunc_calls.append(event)
            yield from asyncio.sleep(0)

        self.e = Event('event')
        self.func = MagicMock(_is_coroutine=False)
        self.cofunc = a
        self.func_listener = EventListener(self.func)
        self.cofunc_listener = EventListener(self.cofunc)

    def test_func(self):
        self.func_listener(self.e)
        self.func.assert_called_once_with(self.e)

    @sync
    @asyncio.coroutine
    def test_cofunc(self):
        yield from self.cofunc_listener(self.e)
        self.assertEqual(self._cofunc_call_count, 1)
        self.assertEqual(self._cofunc_calls[0], self.e)


class TestEventTarget(TestCase):
    def setUp(self):
        self.target = EventTarget()
        self.mock = MagicMock(_is_coroutine=False)
        self.e = Event('click')

    def test_event_dispatch(self):
        self.target.addEventListener('click', self.mock)
        self.assertEqual(len(self.target._listeners), 1)
        self.target.dispatchEvent(self.e)
        self.mock.assert_called_once_with(self.e)

    def test_event_dispatch_empty(self):
        self.target.dispatchEvent(self.e)
        self.mock.assert_not_called()

    def test_event_dispatch_multi(self):
        e1 = Event('click')
        e2 = Event('click')
        self.target.addEventListener('click', self.mock)
        self.target.dispatchEvent(e1)
        self.target.dispatchEvent(e2)
        self.assertEqual(self.mock.call_count, 2)
        self.mock.assert_has_calls([call(e1), call(e2)])

    def test_defferent_event_dispatch(self):
        mock1 = MagicMock(_is_coroutine=False)
        mock2 = MagicMock(_is_coroutine=False)
        e1 = Event('click')
        e2 = Event('event')
        self.target.addEventListener('click', mock1)
        self.target.addEventListener('event', mock2)
        self.assertEqual(len(self.target._listeners), 2)
        self.target.dispatchEvent(e1)
        mock1.assert_called_once_with(e1)
        mock2.assert_not_called()

        self.target.dispatchEvent(e2)
        mock1.assert_called_once_with(e1)
        mock2.assert_called_once_with(e2)

    def test_remove_event(self):
        self.target.addEventListener('click', self.mock)
        self.target.removeEventListener('click', self.mock)
        self.target.dispatchEvent(self.e)
        self.mock.assert_not_called()

    def test_remove_event_multi(self):
        self.target.addEventListener('click', self.mock)
        self.assertEqual(len(self.target._listeners), 1)
        self.target.removeEventListener('click', self.mock)
        self.assertEqual(len(self.target._listeners), 0)
        self.target.dispatchEvent(self.e)
        self.mock.assert_not_called()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import sys
import logging
from pathlib import Path

logger = logging.getLogger('wdom')
CURFILE = Path(__file__).resolve()
CURDIR = CURFILE.parent.resolve()

if __name__ == '__main__':
    sys.path.insert(0, str(CURDIR.parent.resolve()))


def main():
    # from tornado.ioloop import IOLoop
    # from tornado.platform.asyncio import AsyncIOMainLoop
    import asyncio
    from wdom.misc import install_asyncio
    install_asyncio()

    from wdom.server import start_server, get_app, stop_server
    # from wdom.server_aio import start_server, get_app, stop_server
    # from wdom.server_tornado import start_server, get_app, stop_server
    # from wdom.examples.markdown_simple import sample_page
    # from wdom.examples.rev_text import sample_page
    # from wdom.examples.data_binding import sample_page
    # from wdom.examples.todo import sample_page
    from wdom.examples.theming import sample_page
    # from wdom.themes import bootstrap3, mdl, skeleton, pure, semantic, kube, foundation, mui
    from wdom.themes import default
    page = sample_page(default)
    app = get_app(document=page)
    loop = asyncio.get_event_loop()
    server = start_server(app=app, loop=loop)

    try:
        loop.run_forever()
    except KeyboardInterrupt:
        stop_server(server)
        loop.close()


if __name__ == '__main__':
    main()
